/* -------------------------------------------------------------------
AdpKmean_3Dinc.c - Some basic functions for image segmentation

Programmer: Michelle Yan
Revisions:  Paul Hughett                                 

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


   This source file contains some basic functions needed for the   
segmemntation using adaptive K-mean method.                        
                                                                   
  Creat_LOTS:   compute look-up tables for speeding up mean        
                update and interpolation.                          
  Free_LOTS:    free memory occupied by LOTS.                      
  Total_mn:     compute the total mean of each region for given    
                label image.                                       
  Update_Mn:    compute local mean of GRID pixels that belong to a 
                given region and are inside of window.             
  Interp_Mn:    interpolate one value of local mean given region   
                number using trilinear interpolation.              
  Update_Seg1:  update segmentation from one sequence and return   
                total number of voxels whose labels are changed.   
  Update_Seg2:  update segmentation from two sequences having      
                diff. weight on each one and return total number   
                of voxels whose labels are changed.                
  Likelihood_1: compute the likelihood for a given label image     
                of single echo image.                              
  Likelihood_2: compute the likelihood for a given label image     
                of dual-echo image.                                
-------------------------------------------------------------------*/

static const char rcsid[] = "$Id: AdpKmean_3Dinc.c,v 1.13 2000/06/05 17:17:59 hughett Exp $ Release $Name: segm-0-90 $";

/* -------------------------------------------------------------------*/
/* Include proper constants, variables and functions declaration      */
/* -------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "common.h"
#include "definition.h"
#include "errm.h"
#include "image_io.h"
#include "image_inc.h"
#include "memm.h"

#include "AdpKmean_3D.h"

static int   *multBySp_X, *divideBySp_X;
static int   *multBySp_Y, *divideBySp_Y;
static int   *multBySp_Z, *divideBySp_Z;


void
Create_LOTS( int N,		/* Row, Col & Slice No. of image array*/
	     int M,
	     int SliceNo,
	     int Sp_X,		/* Half of window size along X,Y & Z  */
	     int Sp_Y,
	     int Sp_Z )
{
  int i;


  multBySp_X   = (int *) callock(N, sizeof(int));
  divideBySp_X = (int *) callock(N, sizeof(int));
  multBySp_Y   = (int *) callock(M, sizeof(int));
  divideBySp_Y = (int *) callock(M, sizeof(int));
  multBySp_Z   = (int *) callock(SliceNo, sizeof(int));
  divideBySp_Z = (int *) callock(SliceNo, sizeof(int));

  Sp_X = max (1, Sp_X);
  for (i=0; i<N; i++) {
    multBySp_X[i] = i * Sp_X;
    divideBySp_X[i] = i / Sp_X;
  }

  Sp_Y = max (1, Sp_Y);
  for (i=0; i<M; i++) {
    multBySp_Y[i] = i * Sp_Y;
    divideBySp_Y[i] = i / Sp_Y;
  }

  Sp_Z = max (1, Sp_Z);
  for (i=0; i<SliceNo; i++) {
    multBySp_Z[i] = i * Sp_Z;
    divideBySp_Z[i] = i / Sp_Z;
  }
}

/******
FIXME:  There is some bug that manifests only under Linux, as a
segmentation fault when freeing divideBySp_Z.  The pointer itself
seems to be intact.  If no optimization is used, then the exact
numbers of pixels changed per step is the same as the SunOS version,
but the crash happens slightly earlier.  If -O2 optimatization is used,
the number of pixels changed per step is slightly different.  With or
without optimization, the stack seems to be so scrambled that gdb
can not do a successful backtrace.  -PWH 19980804
******/


void
Free_LOTS( void )
{
  free (multBySp_X);
  free (divideBySp_X);
  free (multBySp_Y);
  free (divideBySp_Y);
  free (multBySp_Z);
  free (divideBySp_Z);
  return;
}


/*--------------------------------------------------------------------------*/
/* purpose:                                                                 */
/*   This source code computes total mean of each region based on the given */
/* segmentation.                                                            */
/*--------------------------------------------------------------------------*/

void
Total_Mn( float ***ImgP,	/* Pointers to image                */
	  unsigned char ***SegImgP, /* Pointer to label image           */
	  int N,		/* Number of rows, columes & slices */
	  int M,
	  int SliceNo,
	  float *Means,		/* Pointers to means                */
	  int NoRegion )	/* Expected number of regions       */
{

  float		*ptr_f;
  unsigned char	*ptr_c;
  int			ir, i;

  int				k;

#if 0
  double			*Total_v;
  unsigned int			*Cnts;
#else
  static int			alloced = 0;
  static double			*Total_v = NULL;
  static unsigned int		*Cnts = NULL;
#endif

#if 0
  Total_v = (double *) callock(NoRegion, sizeof(double));
  Cnts = (unsigned int *) callock(NoRegion, sizeof(int));
#else
  if (alloced < NoRegion) {
    if (Total_v) {
      free(Total_v);
      free(Cnts);
    }
    Total_v = (double *) callock(NoRegion, sizeof(double));
    Cnts = (unsigned int *) callock(NoRegion, sizeof(int));
    alloced = NoRegion;
  }
#endif
    
/*--------------------------------------------------------------------------*/
/* Compute mean for each Region                                             */
/*--------------------------------------------------------------------------*/
    for ( ir=0; ir<NoRegion; ir++ ) { 
        Total_v[ir] = 0.0;
	Cnts[ir] = 0;
    }

    for ( k=0; k<SliceNo; k++ ) {
	ptr_c = SegImgP[k][0];
	ptr_f = ImgP[k][0];

	for ( i=N*M; i>0; i-- ) {
	    if ( *ptr_c != BACKGROUND ) {
		ir = *ptr_c - 1;
		Cnts[ir] ++;
		Total_v [ir] += (*ptr_f);
	    }
	    ptr_c++;
	    ptr_f++;
	}

    }

    for ( ir=0; ir<NoRegion; ir++ ) { 
	if ( Cnts[ir] > 0 ) {
	    Means[ir] = Total_v[ir]/Cnts[ir];
	}
	else {
	    Means[ir] = 0.0;
	}
    }

#if 0
    free(Total_v);
    free(Cnts);
#endif
}



/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*   It computes local mean of pixels of every Sp_Z slice that belong*/
/* to same region & are inside of window, and at the grid spacing    */
/* equal to Sp_XY, where window width is equal to (2*GridSp_XY+1,    */
/* 2*GridSp_Z+1) for XY-, Z-axis.                                    */
/*-------------------------------------------------------------------*/

void
Update_Mn( float ****LocMn,
	   float *TotalMn,
	   int NoRegion,	/* Total number of regions             */
	   int Sp_XY,		/* Half of window size along XY & Z    */
	   int Sp_Z,
	   unsigned char ***SegImg,
	   float ***Image, 
	   int N,		/* Row, Col & Slice No. of image array */
	   int M,
	   int SliceNo )
{

    double *sumW;
    float  *ptr_1;
    int    *countW;
    int    Inc_Z, Width;
    int    NL, ML, SL;        /* Row, Cols & Slices of local mean    */
    int    im, jm, km;
    int    i, j, k, iw, jw, kw, ri;

    double *ptrSumW;
    float  *ptrLocMn, *ptrTotalMn;
    int    *ptrCountW;


/*-------------------------------------------------------------------*/
/* Initialization                                                    */
/*-------------------------------------------------------------------*/

    sumW = (double *)callock(NoRegion, sizeof(double));
    countW = (int *) callock(NoRegion, sizeof(int));

    Width = multBySp_Y[1] + 4; 
    Inc_Z = max (Sp_Z, 1);
    SL = divideBySp_Z[SliceNo-1] + 1; /* =(SliceNo-1)/max(Sp_Z,1)+1  */
    NL = divideBySp_X[N-1] + 1;
    ML = divideBySp_Y[M-1] + 1;
    for ( k=0; k<SL; k++ ) { 
	for ( i=0; i<NL; i++ ) { 
	    ptr_1 = LocMn[k][i][0]; 
	    for ( j=0; j<ML*NoRegion; ptr_1++, j++ ) {
		*ptr_1 = 0.0;
	    }
	}
    }

/*-------------------------------------------------------------------*/
/* Compute local mean by averaging with a fix window                 */
/*-------------------------------------------------------------------*/
    for ( km=0, k=-Inc_Z; km<SL; km++ ) {
	k += Inc_Z;
	for ( im=0, i=-Sp_XY; im<NL; im++ ) {
	    i += Sp_XY; 
	    for ( jm=0, j=-Sp_XY; jm<ML; jm++ ) {
		j += Sp_XY;
		if ( SegImg[k][i][j] == BACKGROUND ) {
		    ptrLocMn = LocMn[km][im][jm];
		    for ( ri=0; ri<NoRegion; ri++, ptrLocMn++ ) { 
			(*ptrLocMn) = UNDEFINED;
		    }
		}
		else {
		    ptrCountW = countW;
		    ptrSumW = sumW;
		    for ( ri=0; ri<NoRegion; ri++ ) { 
			*(ptrCountW++) = 0;
			*(ptrSumW++) = 0.0;
		    }

		    for ( kw = max(0, (k-Sp_Z)); 
			  kw < min((k+Sp_Z+1), SliceNo); kw++) {
			for ( iw = max(0, (i-Sp_XY)); 
			      iw < min((i+Sp_XY+1), N); iw++) {
			    for ( jw = max(0, (j-Sp_XY)); 
				  jw < min((j+Sp_XY+1), M); jw++) {
				ptrCountW = countW;
				ptrSumW = sumW;
				for ( ri=0; ri<NoRegion; 
				     ri++, ptrCountW++, ptrSumW++ ) {
				    if (SegImg[kw][iw][jw] == ri+1) {
					(*ptrCountW) ++;
					(*ptrSumW) += Image[kw][iw][jw];
				    }
				}
			    }
			}
		    }
		    ptrCountW = countW;
		    ptrSumW = sumW;
		    ptrLocMn = LocMn[km][im][jm];
		    ptrTotalMn = TotalMn;
		    for ( ri=0; ri<NoRegion; ri++, ptrCountW++, 
			 ptrSumW++, ptrLocMn++, ptrTotalMn++ ) { 
			if ( *ptrCountW < Width ) {
			    *ptrLocMn = *ptrTotalMn;
			}
			else {
			    *ptrLocMn = (*ptrSumW)/(*ptrCountW);
			}
/*			printf("(i, j, k), Mean = (%d, %d, %d); %e\n", 
			       km, im, jm, *ptrLocMn); */
		    }
		}
	    }
	}
    }
    free (sumW);
    free (countW);
}

/*--------------------------------------------------------------------
Interp_Mn - Interpolate to compute local mean

This function interpolates in the local means to find the local mean
for each region at a given point.

This function is moderately optimzed for performance.
----------------------------------------------------------------------*/

void
Interp_Mn( int Pos_xi,         /* Location to interpolate at */
	   int Pos_yj,
	   int Pos_zk,
	   float ****LocMn,    /* 4d array of local means */
	   float *TotalMn,     /* Region means for whole image */
	   int Sp_XY,          /* Region size for local means */
	   int Sp_Z, 
	   unsigned char ***SegImg,   /* Segmented image */
	   int N,              /* Image dimensions */
	   int M,
	   int SliceNo,
	   float *Means,       /* (out) Regional means, by tisse type */
	   int NoRegion )      /* Number of regions */
{  float	*mnp, scl, scale;
   float 	i_ratio, j_ratio, k_ratio;
   int		sg_ful, sg_fur, sg_flr, sg_fll;
   int		sg_bul, sg_bur, sg_blr, sg_bll;
   int		lo_i, up_i, le_j, ri_j, fn_k, bk_k;
   int		lo_im, up_im, le_jm, ri_jm, fn_km, bk_km;
   int		Inc_Z, ireg;

   /* If the position falls exactly on a grid point,
      no interpolation needed */
   Inc_Z = max(Sp_Z, 1); 
   up_im = divideBySp_X[Pos_xi];
   le_jm = divideBySp_Y[Pos_yj];
   bk_km = divideBySp_Z[Pos_zk];

   if ( Pos_xi%Sp_XY == 0 && Pos_yj%Sp_XY == 0 && 
	Pos_zk%Inc_Z == 0 ) {
      for (ireg=0; ireg<NoRegion; ireg++)
	 Means[ireg] = LocMn[bk_km][up_im][le_jm][ireg];
      return;
   }

   /* Otherwise compute the necessary indices */
   up_i = multBySp_X[up_im];
   lo_i = up_i + Sp_XY;
   lo_i = ((lo_i > N-1) ? up_i : lo_i);
   lo_im = divideBySp_X[lo_i];
    
   le_j = multBySp_Y[le_jm];
   ri_j = le_j + Sp_XY;
   ri_j = ((ri_j > M-1) ? le_j : ri_j);
   ri_jm = divideBySp_Y[ri_j];
    
   bk_k = multBySp_Z[bk_km];
   fn_k = bk_k + Inc_Z;
   fn_k = ((fn_k > SliceNo-1) ? bk_k : fn_k);
   fn_km = divideBySp_Z[fn_k];

   /* Find the tissue types at the neighbors */
   sg_ful = SegImg[fn_k][up_i][le_j]; 
   sg_fur = SegImg[fn_k][up_i][ri_j]; 
   sg_flr = SegImg[fn_k][lo_i][ri_j]; 
   sg_fll = SegImg[fn_k][lo_i][le_j]; 
   sg_bul = SegImg[bk_k][up_i][le_j]; 
   sg_bur = SegImg[bk_k][up_i][ri_j]; 
   sg_blr = SegImg[bk_k][lo_i][ri_j]; 
   sg_bll = SegImg[bk_k][lo_i][le_j]; 

   /* If all the neighbors are background, 
      use mean values from the whole image */
   if ( sg_bul == BACKGROUND &&
	sg_bur == BACKGROUND && 
	sg_blr == BACKGROUND &&
	sg_bll == BACKGROUND && 
	sg_ful == BACKGROUND &&
	sg_fur == BACKGROUND && 
	sg_flr == BACKGROUND &&
	sg_fll == BACKGROUND ) {
      for (ireg=0; ireg<NoRegion; ireg++)
	 Means[ireg] = TotalMn[ireg];
      return;
   }

   /* Otherwise prepare for interpolation */
   i_ratio = (Pos_xi - up_i) / (float) Sp_XY;
   j_ratio = (Pos_yj - le_j) / (float) Sp_XY;
   k_ratio = (Pos_zk - bk_k) / (float) Inc_Z;

   scale = 0.0;
   for (ireg=0; ireg<NoRegion; ireg++) 
      Means[ireg] = 0.0;

   /* Average over non-background neighbors */
   if ( sg_bul != BACKGROUND ) {
      scl = (1.0 - i_ratio) * (1.0 - j_ratio) * (1.0 - k_ratio);
      scale += scl;
      mnp = LocMn[bk_km][up_im][le_jm];
      for (ireg=0; ireg<NoRegion; ireg++) 
         Means[ireg] += scl * mnp[ireg];
   }
   if ( sg_bur != BACKGROUND ) {
      scl = (1.0 - i_ratio) * j_ratio * (1.0 - k_ratio);
      scale += scl;
      mnp = LocMn[bk_km][up_im][ri_jm];
      for (ireg=0; ireg<NoRegion; ireg++) 
         Means[ireg] += scl * mnp[ireg];
   }
   if ( sg_blr != BACKGROUND ) {
      scl = i_ratio * j_ratio * (1.0 - k_ratio);
      scale += scl;
      mnp = LocMn[bk_km][lo_im][ri_jm];
      for (ireg=0; ireg<NoRegion; ireg++) 
         Means[ireg] += scl * mnp[ireg];
   }
   if ( sg_bll != BACKGROUND ) {
      scl = i_ratio * (1.0 - j_ratio) * (1.0 - k_ratio);
      scale += scl;
      mnp = LocMn[bk_km][lo_im][le_jm];
      for (ireg=0; ireg<NoRegion; ireg++) 
         Means[ireg] += scl * mnp[ireg];
   }
   if ( sg_ful != BACKGROUND ) {
      scl = (1.0 - i_ratio) * (1.0 - j_ratio) * k_ratio;
      scale += scl;
      mnp = LocMn[fn_km][up_im][le_jm];
      for (ireg=0; ireg<NoRegion; ireg++) 
         Means[ireg] += scl * mnp[ireg];
   }
   if ( sg_fur != BACKGROUND ) {
      scl = (1.0 - i_ratio) * j_ratio * k_ratio;
      scale += scl;
      mnp = LocMn[fn_km][up_im][ri_jm];
      for (ireg=0; ireg<NoRegion; ireg++) 
         Means[ireg] += scl * mnp[ireg];
   }
   if ( sg_flr != BACKGROUND ) {
      scl = i_ratio * j_ratio * k_ratio;
      scale += scl;
      mnp = LocMn[fn_km][lo_im][ri_jm];
      for (ireg=0; ireg<NoRegion; ireg++) 
         Means[ireg] += scl * mnp[ireg];
   }
   if ( sg_fll != BACKGROUND ) {
      scl = i_ratio * (1.0 - j_ratio) * k_ratio;
      scale += scl;
      mnp = LocMn[fn_km][lo_im][le_jm];
      for (ireg=0; ireg<NoRegion; ireg++) 
         Means[ireg] += scl * mnp[ireg];
   }

   for (ireg=0; ireg<NoRegion; ireg++) 
      Means[ireg] /= scale;

   return;
}


/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*   It computes one value of local mean by trilinear interpolation  */
/* based on the result from Update_Mn for pixels without values.     */
/*-------------------------------------------------------------------*/

/* This is an old version kept around just in case */

#if 0

void
old_Interp_Mn( int Pos_xi,	/* Location of interpolation occured   */
	   int Pos_yj,
	   int Pos_zk,
	   float ****LocMn,
	   float *TotalMn,
	   int Sp_XY,
	   int Sp_Z, 
	   unsigned char ***SegImg,
	   int N,
	   int M,
	   int SliceNo,
	   float *Means,
	   int NoRegion )
{

  float		first_mn = 0.0;
  float		mn_ful, mn_fur, mn_fll, mn_flr;
  float		mn_bul, mn_bur, mn_bll, mn_blr;
  double	sc_ful, sc_fur, sc_fll, sc_flr;
  double	sc_bul, sc_bur, sc_bll, sc_blr;
  double	i_ratio, j_ratio, k_ratio;
  int		sg_ful, sg_fur, sg_flr, sg_fll;
  int		sg_bul, sg_bur, sg_blr, sg_bll;
  int		lo_i, up_i, le_j, ri_j, fn_k, bk_k;
  int		lo_im, up_im, le_jm, ri_jm, fn_km, bk_km;
  int		first_in, Inc_Z, Region_I;


    Inc_Z = max(Sp_Z, 1); 
    up_im = divideBySp_X[Pos_xi];
    le_jm = divideBySp_Y[Pos_yj];
    bk_km = divideBySp_Z[Pos_zk];

    if ( Pos_xi%Sp_XY == 0 && Pos_yj%Sp_XY == 0 && 
	 Pos_zk%Inc_Z == 0 ) {
        for (Region_I=0; Region_I<NoRegion; Region_I++)
	  Means[Region_I] = LocMn[bk_km][up_im][le_jm][Region_I];
	return;
    }

    up_i = multBySp_X[up_im];
    lo_i = up_i + Sp_XY;
    lo_i = ((lo_i > N-1) ? up_i : lo_i);
    lo_im = divideBySp_X[lo_i];
    
    le_j = multBySp_Y[le_jm];
    ri_j = le_j + Sp_XY;
    ri_j = ((ri_j > M-1) ? le_j : ri_j);
    ri_jm = divideBySp_Y[ri_j];
    
    bk_k = multBySp_Z[bk_km];
    fn_k = bk_k + Inc_Z;
    fn_k = ((fn_k > SliceNo-1) ? bk_k : fn_k);
    fn_km = divideBySp_Z[fn_k];

    sg_ful = SegImg[fn_k][up_i][le_j]; 
    sg_fur = SegImg[fn_k][up_i][ri_j]; 
    sg_flr = SegImg[fn_k][lo_i][ri_j]; 
    sg_fll = SegImg[fn_k][lo_i][le_j]; 
    sg_bul = SegImg[bk_k][up_i][le_j]; 
    sg_bur = SegImg[bk_k][up_i][ri_j]; 
    sg_blr = SegImg[bk_k][lo_i][ri_j]; 
    sg_bll = SegImg[bk_k][lo_i][le_j]; 

    if (sg_bul == BACKGROUND &&
	sg_bur == BACKGROUND && 
	sg_blr == BACKGROUND &&
	sg_bll == BACKGROUND && 
	sg_ful == BACKGROUND &&
	sg_fur == BACKGROUND && 
	sg_flr == BACKGROUND &&
	sg_fll == BACKGROUND ) {

      for (Region_I=0; Region_I<NoRegion; Region_I++)
	Means[Region_I] = TotalMn[Region_I];
    }
    else {

      i_ratio = (float) (Pos_xi - up_i) / (float) Sp_XY;
      j_ratio = (float) (Pos_yj - le_j) / (float) Sp_XY;
      k_ratio = (float) (Pos_zk - bk_k) / (float) Inc_Z;
	    
#if 0
      sc_bul = (1.0 - i_ratio) * (1.0 - j_ratio) * (1.0 - k_ratio);
      sc_bur = (1.0 - i_ratio) * j_ratio * (1.0 - k_ratio);
      sc_bll = i_ratio * (1.0 - j_ratio) * (1.0 - k_ratio);
      sc_blr = i_ratio * j_ratio * (1.0 - k_ratio);
      sc_ful = (1.0 - i_ratio) * (1.0 - j_ratio) * k_ratio;
      sc_fur = (1.0 - i_ratio) * j_ratio * k_ratio;
      sc_fll = i_ratio * (1.0 - j_ratio) * k_ratio;
      sc_flr = i_ratio * j_ratio * k_ratio;
#endif

      sc_bul = (1.0 - i_ratio) * (1.0 - j_ratio) * (1.0 - k_ratio);
      sc_bur = (1.0 - i_ratio) * (1.0 - k_ratio) - sc_bul;
      sc_bll = (1.0 - j_ratio) * (1.0 - k_ratio) - sc_bul;
      sc_ful = (1.0 - i_ratio) * (1.0 - j_ratio) - sc_bul;
      sc_flr = i_ratio * j_ratio * k_ratio;
      sc_blr = i_ratio * j_ratio - sc_flr;
      sc_fur = j_ratio * k_ratio - sc_flr;
      sc_fll = i_ratio * k_ratio - sc_flr; 

      for (Region_I=0; Region_I<NoRegion; Region_I++) {

	first_in = FALSE;

	mn_ful = LocMn[fn_km][up_im][le_jm][Region_I];
	mn_fur = LocMn[fn_km][up_im][ri_jm][Region_I];
	mn_flr = LocMn[fn_km][lo_im][ri_jm][Region_I];
	mn_fll = LocMn[fn_km][lo_im][le_jm][Region_I];
	mn_bul = LocMn[bk_km][up_im][le_jm][Region_I];
	mn_bur = LocMn[bk_km][up_im][ri_jm][Region_I];
	mn_blr = LocMn[bk_km][lo_im][ri_jm][Region_I];
	mn_bll = LocMn[bk_km][lo_im][le_jm][Region_I];

	if ( sg_bul != BACKGROUND ) {
	    first_in = TRUE;
	    first_mn = mn_bul;
	}
	else if ( sg_bur != BACKGROUND ) {
	    first_in = TRUE;
	    first_mn = mn_bur;
	}
	else if ( sg_blr != BACKGROUND ) {
	    first_in = TRUE;
	    first_mn = mn_blr;
	}
	else if ( sg_bll != BACKGROUND ) {
	    first_in = TRUE;
	    first_mn = mn_bll;
	}
	else if ( sg_ful != BACKGROUND ) {
	    first_in = TRUE;
	    first_mn = mn_ful;
	}
	else if ( sg_fur != BACKGROUND ) {
	    first_in = TRUE;
	    first_mn = mn_fur;
	}
	else if ( sg_flr != BACKGROUND ) {
	    first_in = TRUE;
	    first_mn = mn_flr;
	}
	else if ( sg_fll != BACKGROUND ) {
	    first_in = TRUE;
	    first_mn = mn_fll;
	}
	
	if ( first_in != TRUE ) {
	    (void) printf("Interp_Mn: ALL pixels are BACKGROUND !!!\n");
	}
	else { 
	    if ( sg_bul == BACKGROUND ) {
		mn_bul = first_mn;
	    }
	    if ( sg_bur == BACKGROUND ) {
		mn_bur = first_mn;
	    }
	    if ( sg_blr == BACKGROUND ) {
		mn_blr = first_mn;
	    }
	    if ( sg_bll == BACKGROUND ) {
		mn_bll = first_mn;
	    }
	    if ( sg_ful == BACKGROUND ) {
		mn_ful = first_mn;
	    }
	    if ( sg_fur == BACKGROUND ) {
		mn_fur = first_mn;
	    }
	    if ( sg_flr == BACKGROUND ) {
		mn_flr = first_mn;
	    }
	    if ( sg_fll == BACKGROUND ) {
		mn_fll = first_mn;
	    }

	    Means[Region_I] = sc_bul * mn_bul + sc_bur * mn_bur + 
	                      sc_blr * mn_blr + sc_bll * mn_bll +
			      sc_ful * mn_ful + sc_fur * mn_fur + 
			      sc_flr * mn_flr + sc_fll * mn_fll;
/*	    printf("(%d, %d, %d): interpolated mean = %e\n", 
		   Pos_xi, Pos_yj, Pos_zk, Means[Region_I]); */
	}
    }
  }

}

#endif



/*-------------------------------------------------------------------
Update_Seg1 - Update voxel segmentation to reduce Gibbs energy

This function goes through a single-echo image once, and replaces
the tissue type of each voxel by the tissue type that minimizes an
error consisting of a Gibbs mismatch plus a mismatch between the 
actual gray level and the estimated local true gray level of that 
tissue type.  The number of voxel changed is returned.

The inner loops of this function are optimized for speed rather
than clarity.  Note that err, lik, and Means are allocated 0..255
even though just 1..NoRegion are used; this saves having to test
for index == 0, which happens when a neighboring point is background.
-------------------------------------------------------------------*/

int
Update_Seg1( float ***Image,
	     unsigned char ***SegIdx,
	     int N,		/* Rows, Cols & Slice No of image array */
	     int M,
	     int SliceNo,
	     float ****AdpLocMn, /* Local mean array of K regions        */
	     float *TotalMn, 
	     int Sp_XY,
	     int Sp_Z,
	     int NoRegion,
	     float Gibbs,
	     float InVar,	/* = 1/(2*variance)                     */
	     int Gibbs3D,	/* Indicator to 3D Gibbs potential      */
	     float *Likelihood )
{
    float G1, G1w, G2, G2w;    /* Gibbs 1st/2nd order, weak components */
    float err[256], lik[256];  /* Error and likelihood by tissue type */
    float Means[256];          /* Mean gray level by tissue type */
    float GibbsWeak;           /* Gibbs energy correct for slice spacing */
    float error, min_error;    /* Mean mismatch plus Gibbs mismatch */
    float min_likeli;          /* Mean mismatch alone */
    int	  ChangeNo;            /* Number of voxels changed */
    float tmpval;
    int	  idx_old, min_idx;    /* Tissue type */
    int	  xi, yj, zk, i;
    int	  up_xi, lo_xi, le_yj, ri_yj, fn_zk, bk_zk;
    
    /* Initialize some working variables */
    ChangeNo = 0;
    (*Likelihood) = 0.0;
    min_likeli = 0.0;
    min_idx = 0;

    /* Correct Gibbs energy for voxel spacing ratio in XY/Z */
    /* In particular, use 0.9 for NIA 1.5mm images */
    if ( Gibbs3D == 1 ) 
	GibbsWeak = Gibbs;
    else if ( Gibbs3D == 2 )
	GibbsWeak = 0.9*Gibbs;
    else
	GibbsWeak = 0.6*Gibbs;

    /* Pre-compute Gibbs error assuming all neighbors are wrong */
    G1  = Gibbs;
    G1w = GibbsWeak;
    G2  = 0.707*Gibbs;
    G2w = 0.707*GibbsWeak;
    error = 4*G1 + 4*G2 + 2*G1w + 8*G2w;

    /* Loop over all the voxels in the image */
    for ( zk=0; zk<SliceNo; zk++ ) { 

	fn_zk = zk + 1;
	bk_zk = ((fn_zk == 1) ? 0 : zk - 1);
	fn_zk = ((fn_zk == SliceNo) ? SliceNo-1 : fn_zk);

	for ( xi=0; xi<N; xi++ ) { 
	    lo_xi = xi + 1;
	    up_xi = xi - 1;
		    
	    for ( yj=0; yj<M; yj++ ) { 

		idx_old = SegIdx[zk][xi][yj];
		if ( idx_old != BACKGROUND ) {
		    ri_yj = yj + 1;
		    le_yj = yj - 1;    
		    
		    /* Interp_Mn expects Means to be 1-origin,
		       so we have to fudge it here. */
		    Interp_Mn (xi, yj, zk, AdpLocMn, TotalMn, 
			       Sp_XY, Sp_Z, SegIdx, N, M, SliceNo, 
			       Means+1, NoRegion);

                    /* Add error due to difference from tissue means */
                    for (i = 1; i <= NoRegion; i++) {
                       tmpval = Image[zk][xi][yj] - Means[i];
                       lik[i] = InVar * tmpval * tmpval;
                       err[i] = error + lik[i];
                    }

                    /* Now adjust for any neighbors that are right */
		    /* FIXME:  Bordering err1 and SegIdx should get
                       rid of all the tests and make this even faster. */
                    if (le_yj>=0) err[SegIdx[zk][xi][le_yj]] -= G1;
                    if (ri_yj<M)  err[SegIdx[zk][xi][ri_yj]] -= G1;

                    if (up_xi>=0) {
                       err[SegIdx[zk][up_xi][yj]] -= G1; 
		       if (le_yj>=0) err[SegIdx[zk][up_xi][le_yj]] -= G2;
		       if (ri_yj<M)  err[SegIdx[zk][up_xi][ri_yj]] -= G2;
		    } 

                    if (lo_xi<N)  {
                       err[SegIdx[zk][lo_xi][yj]] -= G1;
		       if (le_yj>=0) err[SegIdx[zk][lo_xi][le_yj]] -= G2;
		       if (ri_yj<M)  err[SegIdx[zk][lo_xi][ri_yj]] -= G2;
		    }

                    if (bk_zk>=0) {
                       err[SegIdx[bk_zk][xi][yj]] -= G1w;
                       if (le_yj>=0) err[SegIdx[bk_zk][xi][le_yj]] -= G2w;
                       if (ri_yj<M)  err[SegIdx[bk_zk][xi][ri_yj]] -= G2w;
                       if (up_xi>=0) err[SegIdx[bk_zk][up_xi][yj]] -= G2w;
                       if (lo_xi<N)  err[SegIdx[bk_zk][lo_xi][yj]] -= G2w;
                    }

                    if (fn_zk<SliceNo) {
                       err[SegIdx[fn_zk][xi][yj]] -= G1w;
                       if (ri_yj<M)  err[SegIdx[fn_zk][xi][ri_yj]] -= G2w;
                       if (le_yj>=0) err[SegIdx[fn_zk][xi][le_yj]] -= G2w;
                       if (up_xi>=0) err[SegIdx[fn_zk][up_xi][yj]] -= G2w;
                       if (lo_xi<N)  err[SegIdx[fn_zk][lo_xi][yj]] -= G2w;
                    }

                    /* Select tissue type with minimum error */
                    min_error = INFINITE;
                    for (i = 1; i <= NoRegion; i++)
                       if (err[i] < min_error)  {
                          min_idx = i;
                          min_error  = err[i]; 
                          min_likeli = lik[i];
                       }

                    /* Make the pixel type equal to the best type */
		    *Likelihood += min_likeli;
		    if ( min_idx != idx_old ) {
		        SegIdx[zk][xi][yj] = (unsigned char) min_idx;
			ChangeNo ++;
		    }
		}
	    }
	}   /* End of loop over voxels */
    }   /* End of conditional on voxel == BACKGROUND */

    return ChangeNo;
}


/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*   It updates segmentation/label, which corresponds to the minimum */
/* error of two values to a set of means.                            */
/*-------------------------------------------------------------------*/

int
Update_Seg2( float ***Img_1,
	     float ***Img_2,
	     unsigned char ***SegIdx,
	     int N,		/* Rows, Cols & Slice No                */
	     int M,
	     int SliceNo,
	     float ****AdpLocMn_1,
	     float ****AdpLocMn_2,
	     float *TotalMn_1,
	     float *TotalMn_2,
	     int Sp_XY,
	     int Sp_Z, 
	     int NoRegion,
	     float Gibbs,
	     float VarW_1GW,	/* = (1/(2*Variance_1))*Weight_1GW      */
	     float VarW_1C,	/* = (1/(2*Variance_1))*Weight_1C       */
	     float VarW_2C,	/* = (1/(2*Variance_2))*Weight_2C       */
	     float VarW_2GW,	/* = (1/(2*Variance_2))*Weight_2GW      */
				/* Weight_1GW + Weight_2GW = 1          */
				/* Weight_1C + Weight_2C = 1            */
	     int Gibbs3D,
	     float *Likelihood,
	     int CSF_Label )	/* (One of) label(s) for CSF (1,2,3...) */
{
  float	tmpval;
  int		idx, idx_old;
  int		xi, yj, zk, i;
  float	error;
  float			img_v1, img_v2, mean_1, mean_2, min_error;
  float			*Means_1, *Means_2;
  float			likeli, min_likeli, GibbsWeak;
  int			up_xi, lo_xi, le_yj, ri_yj, fn_zk, bk_zk;
  int			min_idx, out1_pix, out2_pix, out1_WeakPix, out2_WeakPix;
  int			ChangeNo;


    Means_1=(float *) callock(NoRegion,sizeof(float));
    Means_2=(float *) callock(NoRegion,sizeof(float));
    ChangeNo = 0;
    (*Likelihood) = 0.0;
    min_likeli = 0.0;
    min_idx = 0;

/*-------------------------------------------------------------------*/
/* Account for thickness > in-plane width, 0.15 is for 5mm x 0.7682 (or x 0.8594)  */
/*                                         0.75 is for 3mm x ????    */
/*-------------------------------------------------------------------*/

    GibbsWeak = 0.0;
    if ( Gibbs3D == 1 ) 
        GibbsWeak = Gibbs;
    else if ( Gibbs3D == 6 || Gibbs3D == 5 )
        GibbsWeak = 0.15*Gibbs;
    else if ( Gibbs3D == 3 )
        GibbsWeak = 0.75*Gibbs;
    else {
      fprintf( stderr, "Unexpected value for Gibbs3D in Update_Seg2().\n" );
      fprintf( stderr, "Hard-coded expectations are 1, 3, 5, and 6 for 1-echo images.\n" );
      fprintf( stderr, "Current value is %d.  Please notify programmers.\n", Gibbs3D );
      fprintf( stderr, "mailto:sysadmin@bblmail.psycha.upenn.edu\n" );
      exit( 1 );
    }

    for ( zk=0; zk<SliceNo; zk++ ) { 
        fn_zk = zk + 1;
        bk_zk = ((fn_zk == 1) ? 0 : zk - 1);
        fn_zk = ((fn_zk == SliceNo) ? SliceNo-1 : fn_zk);
        for ( xi=0; xi<N; xi++ ) { 
            for ( yj=0; yj<M; yj++ ) { 

                idx_old = SegIdx[zk][xi][yj];
                if ( idx_old != BACKGROUND ) {
                    img_v1 = Img_1[zk][xi][yj];
                    img_v2 = Img_2[zk][xi][yj];
                    min_error = INFINITE;
                    lo_xi = xi + 1;
                    ri_yj = yj + 1;
                    up_xi = xi - 1;
                    le_yj = yj - 1;    

                    Interp_Mn (xi, yj, zk, AdpLocMn_1, TotalMn_1, 
                               Sp_XY, Sp_Z, SegIdx, N, M, SliceNo, 
                               Means_1, NoRegion);
                    Interp_Mn (xi, yj, zk, AdpLocMn_2, TotalMn_2, 
                               Sp_XY, Sp_Z, SegIdx, N, M, SliceNo, 
                               Means_2, NoRegion);

                    for ( i=0; i<NoRegion; i++ ) {
                        mean_1 = Means_1[i];
                        mean_2 = Means_2[i];
                        idx = i + 1;
/*-------------------------------------------------------------------*/
/*                      CSF corresponds to the label value 2         */
/*-------------------------------------------------------------------*/
                        if ( idx != CSF_Label ) {
			  tmpval = (img_v1 - mean_1);
			  error = VarW_1GW * tmpval * tmpval;
			  tmpval = (img_v2 - mean_2);
			  error += VarW_2GW * tmpval * tmpval;
			} else {
			  tmpval = (img_v1 - mean_1);
			  error = VarW_1C * tmpval * tmpval;
			  tmpval = (img_v2 - mean_2);
			  error += VarW_2C * tmpval * tmpval;
			}
                        likeli = error;
                        
                        out1_pix = 0;
                        out2_pix = 0;
                        out1_WeakPix = 0;
                        out2_WeakPix = 0;
/*-------------------------------------------------------------------*/
/*          compute contribution from 1st-order neighbors            */
/*-------------------------------------------------------------------*/
                        if (SegIdx[zk][up_xi][yj] != idx) {
                            out1_pix ++;
                        }       
                        if (SegIdx[zk][lo_xi][yj] != idx) {
                            out1_pix ++;
                        }       
                        if (SegIdx[zk][xi][ri_yj] != idx) {
                            out1_pix ++;
                        }       
                        if (SegIdx[zk][xi][le_yj] != idx) {
                            out1_pix ++;
                        }       

                        if (SegIdx[bk_zk][xi][yj] != idx) {
                            out1_WeakPix ++;
                        }       
                        if (SegIdx[fn_zk][xi][yj] != idx) {
                            out1_WeakPix ++;
                        }       
/*-------------------------------------------------------------------*/
/*          compute contribution from 2nd-order neighbors            */
/*-------------------------------------------------------------------*/
                        if (SegIdx[zk][up_xi][le_yj] != idx) {
                            out2_pix ++;
                        }
                        if (SegIdx[zk][up_xi][ri_yj] != idx) {
                            out2_pix ++;
                        }
                        if (SegIdx[zk][lo_xi][le_yj] != idx) {
                            out2_pix ++;
                        }
                        if (SegIdx[zk][lo_xi][ri_yj] != idx) {
                            out2_pix ++;
                        }

                        if (SegIdx[bk_zk][up_xi][yj] != idx) {
                            out2_WeakPix ++;
                        }
                        if (SegIdx[bk_zk][lo_xi][yj] != idx) {
                            out2_WeakPix ++;
                        }
                        if (SegIdx[bk_zk][xi][le_yj] != idx) {
                            out2_WeakPix ++;
                        }
                        if (SegIdx[bk_zk][xi][ri_yj] != idx) {
                            out2_WeakPix ++;
                        }
                        if (SegIdx[fn_zk][up_xi][yj] != idx) {
                            out2_WeakPix ++;
                        }
                        if (SegIdx[fn_zk][lo_xi][yj] != idx) {
                            out2_WeakPix ++;
                        }
                        if (SegIdx[fn_zk][xi][ri_yj] != idx) {
                            out2_WeakPix ++;
                        }
                        if (SegIdx[fn_zk][xi][le_yj] != idx) {
                            out2_WeakPix ++;
                        }
                        error += (out1_pix + 0.707*out2_pix)*Gibbs + 
                                 (out1_WeakPix + 0.707*out2_WeakPix)*
                                 GibbsWeak;
                        if ( error < min_error ) {
                            min_error = error;
                            min_idx = idx;
                            min_likeli = likeli;
                        }
                    }
                    SegIdx[zk][xi][yj] = (unsigned char) min_idx;
                    (*Likelihood) += min_likeli;
                    if ( min_idx != idx_old ) {
                        ChangeNo ++;
                    }
                }
            }
        }
    }

    free (Means_1);
    free (Means_2);
    return ChangeNo;
}



/*--------------------------------------------------------------------------*/
/* purpose:                                                                 */
/*   This source code computes the likelihood for one sequence based on the */
/* given segmentation.                                                      */
/*--------------------------------------------------------------------------*/

float
Likelihood_1( float ***ImgP,	/* Pointers to image and mask            */
	      unsigned char ***SegImgP,	/* Pointer to output image               */
	      int N,		/* Number of rows, columes & slices      */
	      int M,
	      int SliceNo,
	      int NoRegion,	/* Expected number of regions            */
	      float InVar )
{

    float          *Means, Likelihood, *ptr_f;
    int            i, ir, k;
    unsigned char  *ptr_c;
    float tmpval;


    Means=(float *) callock(NoRegion,sizeof(float));

/*--------------------------------------------------------------------------*/
/* Compute mean for each Region                                             */
/*--------------------------------------------------------------------------*/
    Total_Mn(ImgP, SegImgP, N, M, SliceNo, Means, NoRegion);

/*--------------------------------------------------------------------------*/
/* Compute the likelihood based on the labels                               */
/*--------------------------------------------------------------------------*/
    Likelihood = 0.0;
    for ( k=0; k<SliceNo; k++ ) {
        ptr_c = SegImgP[k][0];
        ptr_f = ImgP[k][0];
        for ( i=0; i<N*M; i++, ptr_c++, ptr_f++ ) {
            if ( *ptr_c != BACKGROUND ) {
                ir = *ptr_c - 1;
		tmpval = (*ptr_f - Means[ir]);
                Likelihood += tmpval * tmpval;
            }
        }
    }
    Likelihood *= InVar;

    free(Means);

    return Likelihood;
}


/*--------------------------------------------------------------------------*/
/* purpose:                                                                 */
/*   This source code computes the likelihood for two sequence based on the */
/* given segmentation.                                                      */
/*--------------------------------------------------------------------------*/

float
Likelihood_2( float ***ImgP_1,	/* Pointers to image and mask            */
	      float ***ImgP_2,
	      unsigned char ***SegImgP, /* Pointer to output image               */
	      int N,		/* Number of rows, columes & slices      */
	      int M,
	      int SliceNo,
	      int NoRegion,     /* Expected number of regions            */
	      float InVar_1GW,
	      float InVar_2GW,
	      float InVar_1C,
	      float InVar_2C )
{

    float          *Means_1, *Means_2, Likelihood, 
                   Likeli_1GW, Likeli_2GW, Likeli_1C, Likeli_2C, 
                   *ptr_f1, *ptr_f2, t2CSF_V;
    float tmpval;
    int            i, ir, k, csfLabel=NOT_UNINIT;
    unsigned char  *ptr_c;



    Means_1=(float *) callock(NoRegion,sizeof(float));
    Means_2=(float *) callock(NoRegion,sizeof(float));


/*--------------------------------------------------------------------------*/
/* Compute mean for each Region                                             */
/*--------------------------------------------------------------------------*/

    Total_Mn(ImgP_1, SegImgP, N, M, SliceNo, Means_1, NoRegion);
    Total_Mn(ImgP_2, SegImgP, N, M, SliceNo, Means_2, NoRegion);

    t2CSF_V = -1.0;
    csfLabel = -1;
    for ( i=0; i<NoRegion; i++ ) {
      if ( Means_1[i] > t2CSF_V ) {
	t2CSF_V = Means_1[i];
	csfLabel = i;
      } 
    }

/*--------------------------------------------------------------------------*/
/* Compute the likelihood based on the labels                               */
/*--------------------------------------------------------------------------*/
    Likeli_1GW = Likeli_2GW = 0.0;
    Likeli_1C = Likeli_2C = 0.0;
    for ( k=0; k<SliceNo; k++ ) {
	ptr_c = SegImgP[k][0];
	ptr_f1 = ImgP_1[k][0];
	ptr_f2 = ImgP_2[k][0];
	for ( i=0; i<N*M; i++, ptr_c++, ptr_f1++, ptr_f2++) {
	    if ( *ptr_c != BACKGROUND ) {
		ir = *ptr_c - 1;
		if ( ir != csfLabel ) {
		  tmpval = (*ptr_f1 - Means_1[ir]);
		  Likeli_1GW += tmpval * tmpval;
		  tmpval = (*ptr_f2 - Means_2[ir]);
		  Likeli_2GW += tmpval * tmpval;
		}
		else {
		  tmpval = (*ptr_f1 - Means_1[ir]);
		  Likeli_1C += tmpval * tmpval;
		  tmpval = (*ptr_f2 - Means_2[ir]);
		  Likeli_2C += tmpval * tmpval;
		}
	    }
	}
    }
    Likelihood = Likeli_1GW*InVar_1GW + Likeli_2GW*InVar_2GW + 
  	         Likeli_1C*InVar_1C + Likeli_2C*InVar_2C;

    free(Means_1);
    free(Means_2);

    return Likelihood;
}
