/*----------------------------------------------------------------------------
Kmean_3Dinc.c - Some functions to support K-means segmentation

Programmer: Michelle Yan
Revisions:  Paul Hughett, Dan Widyono

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

    This source file contains some basic functions needed for the   
3D segmemntation using K-means method.                              
                                                                    
   minerr_1echo:   return min error of a value to K-mean.           
   minerr_2echo:   return min error of 2 values to 2 K-means.       
   update_Kmean:   update existing means.                           
   split_Kmean:    add one more region.                             
------------------------------------------------------------------------*/

static const char rcsid[] = "$Id: Kmean_3Dinc.c,v 1.9 2000/06/05 17:17:59 hughett Exp $ Release $Name: segm-0-90 $";

/* -------------------------------------------------------------------*/
/* Include proper constants, variables and functions declaration      */
/* -------------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>

#include "definition.h"

#include "Kmean_3Dseg.h"


/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*   It computes the min error of a value to a set of means and      */ 
/* accumulates contribution to sum and count, where a 3D second order*/
/* neighborhood is used for a prior.                                 */
/*-------------------------------------------------------------------*/

float
minerr_1echo( float ImgValue,
	      int Pos_xi,
	      int Pos_yj,
	      int Pos_zk,
	      unsigned char ***SegIndex,
	      int N,		/* Rows, Cols & Slices of 3D array      */
	      int M, 
	      int L,
	      Kmean_Info K_Mean,
	      int NoRegion,
	      float GibbsPotential,
	      float Weight, /* 1/(2*variance) */
	      int Gibbs3D )
{

  register float    error, min_error = INFINITE;
  int      index; 
  int      upper_xi=NOT_UNINIT, lower_xi=NOT_UNINIT, 
    left_yj=NOT_UNINIT, right_yj=NOT_UNINIT, 
    front_zk=NOT_UNINIT, back_zk=NOT_UNINIT;
  int      i, min_index=NOT_UNINIT, out1_pixel, out2_pixel;
  float	   tmpval;


  if (GibbsPotential) {
    lower_xi = Pos_xi + 1;
    right_yj = Pos_yj + 1;
    front_zk = Pos_zk + 1;
    upper_xi = Pos_xi - 1;
    left_yj = Pos_yj - 1;
    back_zk = Pos_zk - 1;
    /*  Bounds check  */
    if (upper_xi < 0)
      upper_xi = 0;
    else if (lower_xi == N)
      lower_xi = N-1;
    if (left_yj < 0)
      left_yj = 0;
    else if (right_yj == M)
      right_yj = M-1;
    if (back_zk < 0)
      back_zk = 0;
    else if (front_zk == L)
      front_zk = L-1;
  }

#if 0
  upper_xi = ((lower_xi == 1) ? 0 : Pos_xi - 1);
  left_yj = ((right_yj == 1) ? 0 : Pos_yj - 1);    
  back_zk = ((front_zk == 1) ? 0 : Pos_zk - 1);
  lower_xi = ((lower_xi == N) ? N-1 : lower_xi);
  right_yj = ((right_yj == M) ? M-1 : right_yj);
  front_zk = ((front_zk == L) ? L-1 : front_zk);
#endif
    
  for ( i=0; i<NoRegion; i++ ) {
    tmpval = ImgValue - K_Mean.K_Mean[i];
    error = Weight * tmpval * tmpval;
	
    if ( GibbsPotential != 0.0 ) { 
      index = i + 1;
      out1_pixel = 0;
      out2_pixel = 0;

/*-------------------------------------------------------------------*/
/*          compute contribution from 1st-order neighbors            */
/*-------------------------------------------------------------------*/
      if ( SegIndex[Pos_zk][upper_xi][Pos_yj] != index ) {
	out1_pixel ++;
      }	
      if ( SegIndex[Pos_zk][lower_xi][Pos_yj] != index ) {
	out1_pixel ++;
      }	
      if ( SegIndex[Pos_zk][Pos_xi][right_yj] != index ) {
	out1_pixel ++;
      }	
      if ( SegIndex[Pos_zk][Pos_xi][left_yj] != index ) {
	out1_pixel ++;
      }	
      if ( Gibbs3D == TRUE ) { 
	if ( SegIndex[back_zk][Pos_xi][Pos_yj] != index ) {
	  out1_pixel ++;
	}	
	if ( SegIndex[front_zk][Pos_xi][Pos_yj] != index ) {
	  out1_pixel ++;
	}	
      }

/*-------------------------------------------------------------------*/
/*          compute contribution from 2nd-order neighbors            */
/*-------------------------------------------------------------------*/
      if ( SegIndex[Pos_zk][upper_xi][left_yj] != index ) {
	out2_pixel ++;
      }
      if ( SegIndex[Pos_zk][upper_xi][right_yj] != index ) {
	out2_pixel ++;
      }
      if ( SegIndex[Pos_zk][lower_xi][left_yj] != index ) {
	out2_pixel ++;
      }
      if ( SegIndex[Pos_zk][lower_xi][right_yj] != index ) {
	out2_pixel ++;
      }

      if ( Gibbs3D == TRUE ) {
	if ( SegIndex[back_zk][upper_xi][Pos_yj] != index ) {
	  out2_pixel ++;
	}
	if ( SegIndex[front_zk][upper_xi][Pos_yj] != index ) {
	  out2_pixel ++;
	}
	if ( SegIndex[back_zk][Pos_xi][left_yj] != index ) {
	  out2_pixel ++;
	}
	if ( SegIndex[back_zk][Pos_xi][right_yj] != index ) {
	  out2_pixel ++;
	}
	if ( SegIndex[front_zk][Pos_xi][right_yj] != index ) {
	  out2_pixel ++;
	}
	if ( SegIndex[front_zk][Pos_xi][left_yj] != index ) {
	  out2_pixel ++;
	}
	if ( SegIndex[back_zk][lower_xi][Pos_yj] != index ) {
	  out2_pixel ++;
	}
	if ( SegIndex[front_zk][lower_xi][Pos_yj] != index ) {
	  out2_pixel ++;
	}
      }
      error += (out1_pixel + 0.707*out2_pixel) * GibbsPotential;
    } /* if (GibbsPotential > 0) */

    if ( error < min_error ) {
      min_error = error;
      min_index = i;
    }
  }
    
  SegIndex[Pos_zk][Pos_xi][Pos_yj] = min_index + 1;
  (K_Mean.Kmean_Count)[min_index] ++;
  (K_Mean.Kmean_Sum)[min_index] += ImgValue;

  return min_error;
}



/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*   It computes the min error of two values to their corresponding  */
/* sets and accumulates contribution to sum and count.               */
/*-------------------------------------------------------------------*/

float
minerr_2echo( float ImgValue_1,
	      float ImgValue_2,
	      int Pos_xi,
	      int Pos_yj,
	      int Pos_zk, 
	      float Weight_1, /* = 1/(2*Variance_1)                   */
	      float Weight_2, /* = 1/(2*Variance_1)                   */
	      unsigned char ***SegIndex,
	      int N,		/* Rows, Cols & Slices of 3D array      */
	      int M,
	      int L,
	      Kmean_Info Kmean_1, 
	      Kmean_Info Kmean_2,
	      int NoRegion,
	      float GibbsPotential,
	      int Gibbs3D )
{

    float    error, min_error = INFINITE;
    float    tmpval1, tmpval2;
    int      index; 
    int      upper_xi, lower_xi, left_yj, right_yj, 
             front_zk, back_zk;
    int      i, min_index=NOT_UNINIT, out1_pixel, out2_pixel;

    lower_xi = Pos_xi + 1;
    right_yj = Pos_yj + 1;
    front_zk = Pos_zk + 1;
    upper_xi = ((lower_xi == 1) ? 0 : Pos_xi - 1);
    left_yj = ((right_yj == 1) ? 0 : Pos_yj - 1);    
    back_zk = ((front_zk == 1) ? 0 : Pos_zk - 1);
    lower_xi = ((lower_xi == N) ? N-1 : lower_xi);
    right_yj = ((right_yj == M) ? M-1 : right_yj);
    front_zk = ((front_zk == L) ? L-1 : front_zk);

    for ( i=0; i<NoRegion; i++ ) {
      tmpval1 = ImgValue_1 - Kmean_1.K_Mean[i];
      tmpval2 = ImgValue_2 - Kmean_2.K_Mean[i];
      error =
	Weight_1 * tmpval1 * tmpval1 +
	Weight_2 * tmpval2 * tmpval2;

      if ( GibbsPotential != 0.0 ) { 
	index = i + 1;
	out1_pixel = 0;
	out2_pixel = 0;

/*-------------------------------------------------------------------*/
/*          compute contribution from 1st-order neighbors            */
/*-------------------------------------------------------------------*/
	if ( SegIndex[Pos_zk][upper_xi][Pos_yj] != index ) {
	  out1_pixel ++;
	}	
	if ( SegIndex[Pos_zk][lower_xi][Pos_yj] != index ) {
	  out1_pixel ++;
	}	
	if ( SegIndex[Pos_zk][Pos_xi][right_yj] != index ) {
	  out1_pixel ++;
	}	
	if ( SegIndex[Pos_zk][Pos_xi][left_yj] != index ) {
	  out1_pixel ++;
	}	
	if ( Gibbs3D == TRUE ) {
	  if ( SegIndex[back_zk][Pos_xi][Pos_yj] != index ) {
	    out1_pixel ++;
	  }	
	  if ( SegIndex[front_zk][Pos_xi][Pos_yj] != index ) {
	    out1_pixel ++;
	  }	
	}

/*-------------------------------------------------------------------*/
/*          compute contribution from 2nd-order neighbors            */
/*-------------------------------------------------------------------*/

	if ( SegIndex[Pos_zk][upper_xi][left_yj] != index ) {
	  out2_pixel ++;
	}
	if ( SegIndex[Pos_zk][upper_xi][right_yj] != index ) {
	  out2_pixel ++;
	}
	if ( SegIndex[Pos_zk][lower_xi][left_yj] != index ) {
	  out2_pixel ++;
	}
	if ( SegIndex[Pos_zk][lower_xi][right_yj] != index ) {
	  out2_pixel ++;
	}
	    
	if ( Gibbs3D == TRUE ) {
	  if ( SegIndex[back_zk][upper_xi][Pos_yj] != index ) {
	    out2_pixel ++;
	  }
	  if ( SegIndex[front_zk][upper_xi][Pos_yj] != index ) {
	    out2_pixel ++;
	  }
	  if ( SegIndex[back_zk][Pos_xi][left_yj] != index ) {
	    out2_pixel ++;
	  }
	  if ( SegIndex[back_zk][Pos_xi][right_yj] != index ) {
	    out2_pixel ++;
	  }
	  if ( SegIndex[front_zk][Pos_xi][right_yj] != index ) {
	    out2_pixel ++;
	  }
	  if ( SegIndex[front_zk][Pos_xi][left_yj] != index ) {
	    out2_pixel ++;
	  }
	  if ( SegIndex[back_zk][lower_xi][Pos_yj] != index ) {
	    out2_pixel ++;
	  }
	  if ( SegIndex[front_zk][lower_xi][Pos_yj] != index ) {
	    out2_pixel ++;
	  }
	}
	error += (out1_pixel + 0.707*out2_pixel) * GibbsPotential;
      }

      if ( error < min_error ) {
	min_error = error;
	min_index = i;
      }
    }
    
    SegIndex[Pos_zk][Pos_xi][Pos_yj] = min_index + 1;

    (Kmean_1.Kmean_Count)[min_index] ++;
    (Kmean_1.Kmean_Sum)[min_index] += ImgValue_1;
    (Kmean_2.Kmean_Count)[min_index] ++;
    (Kmean_2.Kmean_Sum)[min_index] += ImgValue_2;

    return min_error;
}


/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*   It updates the existing means.                                  */
/*-------------------------------------------------------------------*/

void
update_Kmean( Kmean_Info K_Mean,
	      int NoRegion,
	      float Delta )	/* A perturbation used in splitting regions  */
{

    int     i;

    for ( i=0; i<NoRegion; i++ ) {
	printf("update_Kmean: Kmean_Count[%d] = %d\n", i+1, 
	       K_Mean.Kmean_Count[i]);
	if ( (K_Mean.Kmean_Count)[i] > 0 ) 
	    K_Mean.K_Mean[i] = K_Mean.Kmean_Sum[i]/
		                K_Mean.Kmean_Count[i];
	else
	    K_Mean.K_Mean[i] = (1.0 + Delta)*K_Mean.Total_Mean;
    }

}


/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*   It split the K_means (add one more region).                     */ 
/*-------------------------------------------------------------------*/

void
split_Kmean( Kmean_Info K_Mean,
	     int NoRegion,
	     float Delta )	/* A perturbation used in splitting regions  */
{
    K_Mean.K_Mean[NoRegion-1] = (1.0 + Delta)*K_Mean.K_Mean[0];
}



