/*--------------------------------------------------------------------------
Kmean_3Dseg.c                                                            

Programmer: Michelle Yan
Revisions:  Paul Hughett, Dan Widyono

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

   It segments an image using K-mean (ML or MAP) method. If interested in 
only brain volume, BrainMask should be used prior.                     

Including:                                                         
         Kmean3D_1echo();                                                
         Kmean3D_2echo();                                                
--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
Some pre-RCS history:
  DRW 19980131: optimization
  DRW 19980225: kept interactive code in main(), placed rest in
     separate function which can in the future be called by Tcl/Tk
     or some other external program.
  PWH 19980310: Ported to Linux, fixed gcc warnings and errors.
-----------------------------------------------------------------------------*/

static const char rcsid[] = "$Id: Kmean_3Dseg.c,v 1.17 2000/06/05 17:17:59 hughett Exp $ Release $Name: segm-0-90 $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "common.h"
#include "definition.h"
#include "errm.h"
#include "image_io.h"
#include "memm.h"

#include "Kmean_3Dseg.h"


/*--------------------------------------------------------------------------
Kmean3D_1echo - segment a single-echo MRI volume image of the brain

This source code segments an image corresponging to non-background into
several regions, whose number is no more than a given RegionNo.

FIXME:  The arguments to this function need to be understood and
documented.
--------------------------------------------------------------------------*/

static void
Kmean3D_1echo( 
    float ***ImgP,	      /* Pointers to image and mask     */
    float ***LabelP,
    unsigned char ***SegImgP, /* Pointer to output image        */
    int N,		      /* Number of rows, cols & slices  */
    int M,
    int L,
    int NoRegion,	      /* Expected number of regions     */
    float Perturb,	      /* Perturb. for splitting regions */
    float Epsilon,	      /* Distortion threshold of means  */
    float GibbsPotential,
    float Weight,
    Kmean_Info Kmean,
    int Gibbs3D )
{

  double        		err_old, err_new;
  float				calc_err, min_err;
  float		tmpval;
  float         		threshold;
  float         		*Ptr_1, *Ptr_2;
  int				total_region, k;
  int  		curr_region, j, i;
  float		*ImgPtr, *LabelPtr;
  unsigned char	*SegImgPtr;
  int				min_err_index=NOT_UNINIT, index_old;
  int           		change_no, MinChange, IterNo;
  unsigned char 		*Ptr_3;


/*--------------------------------------------------------------------------*/
/* Initialization                                                           */
/*--------------------------------------------------------------------------*/

  MinChange = L * max(N, M)/10.;
  for ( k=0; k<L; k++ ) {
    Ptr_2 = LabelP[k][0];
    Ptr_3 = SegImgP[k][0];
    for (i = N*M; i > 0; i-- ) {
      *Ptr_3++ = (*Ptr_2++ == BACKGROUND ? BACKGROUND : 1);
    }
  }

/*--------------------------------------------------------------------------*/
/* Compute Total mean                                                       */
/*--------------------------------------------------------------------------*/

  Kmean.Kmean_Sum[0] = 0.0;
  Kmean.Kmean_Count[0] = 0;

  for ( k=0; k<L; k++ ) {
    Ptr_1 = ImgP[k][0];
    Ptr_2 = LabelP[k][0];

    for ( i=N*M; i>0; i-- ) {
      if ( *Ptr_2++ != BACKGROUND ) {
	Kmean.Kmean_Sum[0] += *Ptr_1;
	Kmean.Kmean_Count[0] ++;
      }
      Ptr_1++;
    }
  }
  if ( Kmean.Kmean_Count[0] == 0 ) 
    fatal("Kmean2D_1echo: zero image.");
  else {
    Kmean.Total_Mean = Kmean.Kmean_Sum[0] / Kmean.Kmean_Count[0];
    Kmean.K_Mean[0] = Kmean.Total_Mean;
  }
  printf("Kmean3D_1echo: Total mean = %f\n", Kmean.Total_Mean);


/*--------------------------------------------------------------------------*/
/* Find K-mean segmentation by starting with 2 guessed means                */
/*--------------------------------------------------------------------------*/

  total_region = 1;
  while ( total_region < NoRegion ) {
    err_new = INFINITE;

    total_region ++;
    /*    split_Kmean ( Kmean, total_region, Perturb ); */
    Kmean.K_Mean[total_region - 1] = (1.0 + Perturb) * Kmean.K_Mean[0];

    IterNo = 0;
    do {
      IterNo ++;

      err_old = err_new;
      for ( i=0; i<total_region; i++ ) {
	printf("Kmean3D_1echo: K_Mean[%d] = %0.0f\n", i+1, 
	       Kmean.K_Mean[i]);
      }
      do {
	err_new = 0.0;
	for ( i=0; i<total_region; i++ ) {
	  Kmean.Kmean_Sum[i] = 0.0;
	  Kmean.Kmean_Count[i] = 0;
	}

	change_no = 0;

	for ( k=0; k<L; k++ ) {

	  /* Array[k][i][j] */
	  ImgPtr = ImgP[k][0];
	  SegImgPtr = SegImgP[k][0];
	  LabelPtr = LabelP[k][0];

	  for ( i=0; i<N; i++ ) {

	    for ( j=0; j<M; j++ ) {

	      if ( *LabelPtr != BACKGROUND ) {

		index_old = *SegImgPtr;

		if (GibbsPotential) {
		  err_new += minerr_1echo (*ImgPtr, i, j, k,
					   SegImgP, N, M, L, Kmean,
					   total_region, GibbsPotential, 
					   Weight, Gibbs3D);
		} else {

		  /*  Inlining GibbsPotential=0 special case cuts
		   *  almost 1/3 time.  */
		  min_err = INFINITE;

		  for (curr_region = total_region - 1;
		       curr_region >= 0;
		       curr_region--) {

		    /*  Eliminating pow() function call cuts > 2/3 time */

		    tmpval = *ImgPtr - Kmean.K_Mean[curr_region];
		    calc_err = tmpval * tmpval * Weight;

		    if (calc_err < min_err) {
		      min_err = calc_err;
		      min_err_index = curr_region;
		    }
		  }

		  *SegImgPtr = min_err_index + 1;
		  (Kmean.Kmean_Count)[min_err_index] ++;
		  (Kmean.Kmean_Sum)[min_err_index] += *ImgPtr;
		  err_new += min_err;

		}

		if ( index_old != *SegImgPtr ) {
		  change_no ++;
		}

	      }

	      LabelPtr++;
	      SegImgPtr++;
	      ImgPtr++;

	    } /* for j */

	  } /* for i */

	} /* for k */

	printf("No. of Pixels changed = %d\n", change_no);

      } while ( GibbsPotential != 0. && change_no > MinChange );

      threshold = (err_old - err_new )/err_new;
      printf("Iter. %d: err_old, err_new, threshold = %0.3e %0.3e %0.3e\n",
	     IterNo, err_old, err_new, threshold);
      if ( threshold > Epsilon ) {
	update_Kmean ( Kmean, total_region, Perturb );
      }

    } while ( threshold > Epsilon && IterNo < 20);

  }
}	    



/*--------------------------------------------------------------------------
Kmean3D_2echo - segment a dual-echo volume image of the brain

This source code segments a dual-echo iamge corresponding to
non-background into several regions, whose number is no more than a
given RegionNo.

FIXME:  The arguments to this function need to be understood and
documented.
--------------------------------------------------------------------------*/

static void
Kmean3D_2echo( 
    float ***ImgP_1,	      /* Pointers to images and their mask */
    float ***ImgP_2,
    float ***LabelP,
    unsigned char ***SegImgP, /* Pointer to output image */
    int N,		      /* Number of rows, cols & slices */
    int M,
    int L,
    int NoRegion, 	      /* Expected number of regions */
    float Perturb,	      /* Perturb. for splitting regions */
    float Epsilon,	      /* Distortion threshold of means */
    float GibbsPotential,
    Kmean_Info Kmean_1,
    Kmean_Info Kmean_2, 
    float weight_1,
    float weight_2,
    int Gibbs3D )
{

  double			err_old, err_new;
  float				calc_err, min_err;
  float		tmpval;
  float				threshold;
  float				*Ptr_1, *Ptr_2, *Ptr_3;
  int				total_region, k;
  int			curr_region, j, i;
  float		*ImgPtr_1, *ImgPtr_2, *LabelPtr;
  unsigned char	*SegImgPtr;
  int				min_err_index=NOT_UNINIT, index_old;
  int				change_no, MinChange, IterNo;
  unsigned char			*Ptr_4;


/*--------------------------------------------------------------------------*/
/* Initialization                                                           */
/*--------------------------------------------------------------------------*/

  MinChange = L * max(N, M)/10.;
  for ( k=0; k<L; k++ ) {
    Ptr_2 = LabelP[k][0];
    Ptr_4 = SegImgP[k][0];
    for (i = N*M; i > 0; i--) {
      *Ptr_4++ = (*Ptr_2++ == BACKGROUND ? BACKGROUND : 1);
    }
  }

/*--------------------------------------------------------------------------*/
/* Compute Total means                                                      */
/*--------------------------------------------------------------------------*/

  Kmean_1.Kmean_Sum[0] = 0.0;
  Kmean_1.Kmean_Count[0] = 0;
  Kmean_2.Kmean_Sum[0] = 0.0;
  Kmean_2.Kmean_Count[0] = 0;

  for ( k=0; k<L; k++) {
    Ptr_1 = ImgP_1[k][0];
    Ptr_2 = ImgP_2[k][0];
    Ptr_3 = LabelP[k][0];

    for ( i=N*M; i > 0; i-- ) {
      if ( *Ptr_3++ != BACKGROUND ) {
	Kmean_1.Kmean_Sum[0] += *Ptr_1;
	Kmean_1.Kmean_Count[0] ++;
	Kmean_2.Kmean_Sum[0] += *Ptr_2;
	Kmean_2.Kmean_Count[0] ++;
      }
      Ptr_1++;
      Ptr_2++;
    }
  }
  if ( Kmean_1.Kmean_Count[0] == 0 || Kmean_2.Kmean_Count[0] == 0 ) 
    fatal("Kmean3D_2echo: zero image.");
  else {
    Kmean_1.Total_Mean = Kmean_1.Kmean_Sum[0] / Kmean_1.Kmean_Count[0];
    Kmean_2.Total_Mean = Kmean_2.Kmean_Sum[0] / Kmean_2.Kmean_Count[0];
    Kmean_1.K_Mean[0] = Kmean_1.Total_Mean;
    Kmean_2.K_Mean[0] = Kmean_2.Total_Mean;
  }
  printf("Kmean3D_2echo: TotalMean_1, TotalMean_2 = %f, %f\n", 
	 Kmean_1.Total_Mean, Kmean_2.Total_Mean);


/*--------------------------------------------------------------------------*/
/* Find K-mean segmentation by starting with 2 guessed means                */
/*--------------------------------------------------------------------------*/

  total_region = 1;
  while ( total_region < NoRegion ) {
    err_new = INFINITE;

    total_region ++;
    split_Kmean ( Kmean_1, total_region, Perturb );
    split_Kmean ( Kmean_2, total_region, Perturb );

    IterNo = 0;
    do {
      IterNo ++;

      err_old = err_new;
      for ( i=0; i<total_region; i++ ) {
	printf("Kmean3D_2echo: Kmean_1[%d],Kmean_2[%d] = %f,%f\n", 
	       i+1, i+1, Kmean_1.K_Mean[i], Kmean_2.K_Mean[i]);
      }
      do { 
	err_new = 0.0;
	for ( i=0; i<total_region; i++ ) {
	  Kmean_1.Kmean_Sum[i] = 0.0;
	  Kmean_1.Kmean_Count[i] = 0;
	  Kmean_2.Kmean_Sum[i] = 0.0;
	  Kmean_2.Kmean_Count[i] = 0;
	}

	change_no = 0;		

	for ( k=0; k<L; k++ ) { 

	  /* Array[k][i][j] */
	  ImgPtr_1 = ImgP_1[k][0];
	  ImgPtr_2 = ImgP_2[k][0];
	  SegImgPtr = SegImgP[k][0];
	  LabelPtr = LabelP[k][0];

	  for ( i=0; i<N; i++ ) {

	    for ( j=0; j<M; j++ ) {

	      if ( *LabelPtr != BACKGROUND ) {

		index_old = *SegImgPtr;

		if (GibbsPotential) {
		  err_new += minerr_2echo (*ImgPtr_1, *ImgPtr_2, i, j, k, 
					   weight_1, weight_2, SegImgP, 
					   N, M, L, Kmean_1, Kmean_2, 
					   total_region, GibbsPotential, 
					   Gibbs3D);
		} else {

		  /*  Inlining GibbsPotential=0 special case cuts
		   *  almost 1/3 time.  */
		  min_err = INFINITE;

		  for (curr_region = total_region - 1;
		       curr_region >= 0;
		       curr_region--) {

		    /*  Eliminating pow() function call cuts > 2/3 time */

		    tmpval = *ImgPtr_1 - Kmean_1.K_Mean[curr_region];
		    calc_err = tmpval * tmpval * weight_1;
		    tmpval = *ImgPtr_2 - Kmean_2.K_Mean[curr_region];
		    calc_err += tmpval * tmpval * weight_2;

		    if (calc_err < min_err) {
		      min_err = calc_err;
		      min_err_index = curr_region;
		    }
		  }

		  *SegImgPtr = min_err_index + 1;
		  (Kmean_1.Kmean_Count)[min_err_index] ++;
		  (Kmean_1.Kmean_Sum)[min_err_index] += *ImgPtr_1;
		  (Kmean_2.Kmean_Count)[min_err_index] ++;
		  (Kmean_2.Kmean_Sum)[min_err_index] += *ImgPtr_2;
		  err_new += min_err;

		}

		if ( index_old != *SegImgPtr ) {
		  change_no ++;
		}

	      }

	      LabelPtr++;
	      SegImgPtr++;
	      ImgPtr_1++;
	      ImgPtr_2++;

	    } /* for j */

	  } /* for i */

	} /* for k */

	printf("No. of Pixels changed = %d\n", change_no);

      } while ( GibbsPotential != 0. && change_no > MinChange );

      threshold = (err_old - err_new )/err_new;
      printf("Iter. %d: err_old, err_new, threshold = %e %e %e\n",
	     IterNo, err_old, err_new, threshold);
      if ( threshold > Epsilon ) {
	update_Kmean ( Kmean_1, total_region, Perturb );
	update_Kmean ( Kmean_2, total_region, Perturb );
      }

    } while ( threshold > Epsilon && IterNo < 20);
  }
}	    


/*--------------------------------------------------------------------
kmean_3dseg - Do 3D brain segmentation, including file i/o

This function is called by the kmeans and Kmeans_3Dseg programs
to do the actual file input/output and image segmentation once
all the parameters have been determined.

Its operation is not yet completely understood, but here are some
comments:

If First is set to zero, it is interpreted to mean the first slice
which containts at least one non-zero pixel.  If Last is set to zero,
it is interpreted to mean the last slice after First which contains at
least one non-zero pixel.

FIXME: Res_XY and Res_Z are used to choose tuning parameters for the
segmentation algorithms, which is NOT a reasonable and portable
method.

FIXME:  Several of the function arguments are not well understood.

--------------------------------------------------------------------*/

int
kmean_3dseg(
    int echo_12,           /* Number of echoes: 1 or 2 */
    int NoRegion,          /* Number of tissue types to find */
    const char *Name1,     /* Filename of the first image */
    int ImgType1,          /* Type of the first image */
    float Weight_1,        /* ??? */
    const char *Name2,     /* Filename of the second image */
    int ImgType2,          /* Type of the second image */
    float Weight_2,        /* ??? */
    int N,                 /* Number of rows */
    int M,                 /* Number of columns */
    int First,             /* First slice to process, or zero */
    int Last,              /* Last slice to process, or zero */
    int Total,             /* Number of slices */
    float Res_XY,          /* Voxel size in-plane [mm] */
    float Res_Z,           /* Voxel spacing between planes [mm] */
    float Perturb,         /* ??? */
    float Epsilon,         /* ??? */
    float GibbsPotential,  /* ??? */
    int savemean )         /* Save the means image? */
{
  char		*ptr; 
  FILE          *File_1, *File_2=0, *File_Out, *File_Mean=0;
  float         ***ImageP_1=0, ***ImageP_2=0, ***LabelP, **TmpImgP;
  float         *ptr_1, *ptr_2;
  int           Gibbs3D = FALSE;
  int           i, k, ir;
  int           Intensity_v, *NoVoxel;
  unsigned char ***SegImgP, *ptr_c;
  int		SliceNo = 0;
  int		retval = 0;

  Kmean_Info    Kmean_1, Kmean_2;
    
  int		FirstMarker = 0, LastMarker = 0;



  if ( Res_Z/Res_XY < 1.2 ) {
    Gibbs3D = TRUE;
  }
  if ( Gibbs3D == TRUE && GibbsPotential != 0.0 ) 
    printf("True 3D Gibbs potential used.\n");
  else if ( Gibbs3D == FALSE && GibbsPotential != 0.0 ) 
    printf("Weak 3D Gibbs potential applied.\n");

  printf("\n\n");


/*--------------------------------------------------------------------------*/
/*Open an input and output file for reading and writing                     */
/*--------------------------------------------------------------------------*/

  ptr = callock( (strlen(Name1)+5), sizeof(char));

  /*  File_Out is the output from the Most Likely segmentation step,
   *  to be read by the Adaptive K-means step.  */
  strcpy (ptr, "ml_");
  strcat (ptr, Name1);
  if (!(File_Out = open_file(ptr, "w"))) {
    retval = -1;
    goto close_none;
  }

  printf( "savemean is %d\n", savemean );
  if (savemean) {
    if (!(File_Mean = open_file("Mean.img", "w"))) {
      retval = -1;
      goto close_out;
    }
  }

  if (!(File_1 = open_file(Name1, "r"))) {
    retval = -1;
    goto close_mean;
  }

  if (echo_12 == 2 && !(File_2 = open_file(Name2, "r"))) {
    retval = -1;
    goto close_one;
  }


  TmpImgP = alloc_img(N, M);

/*--------------------------------------------------------------------------*/
/*Skip image slices from 1 to (FirstSlice - 1)                              */
/*--------------------------------------------------------------------------*/

  FirstMarker = (First ? First : Total);
  for (k = 1; k < FirstMarker; k++) {
    fread( (void *) TmpImgP[0],
	   N * M * abs(ImgType1), 1,
	   File_1 );
    if (!First) {
      for (i = N * M - 1; i >= 0; i--) {
	if (*(TmpImgP[0] + i)) {
	  /* Oops, we have hit the first slice with brain; rewind to
             its start */
	  fseek( File_1, -(N * M * abs(ImgType1)), SEEK_CUR );
	  First = k;
	  k = FirstMarker;
	  printf( "First slice: %d\n", First );
	  break;
	}
      }
      if (savemean) {
	fwrite( (const void *) TmpImgP[0],
		N * M * abs(ImgType1), 1,
		File_Mean );
      }
    }
    else if (savemean) {
      fwrite( (const void *) TmpImgP[0],
	      N * M * abs(ImgType1), 1,
	      File_Mean );
    }
  }
  if (echo_12 == 2) {
    fseek( File_2,
	   N * M * (First - 1) * abs(ImgType2),
	   SEEK_SET );
  }
  memset( (void *) TmpImgP[0], 0, N * M );
  for (k = 1; k < First; k++) {
    fwrite( (const void *) TmpImgP[0],
	    N * M, 1,
	    File_Out );
  }



  /*
   * Since we may not know how many slices we're going to have yet,
   * let us allocate enough room for pointers for the rest of the slices.
   */

  if (!Last) {
    ImageP_1 = (float ***) callock( Total - First + 1, sizeof( void * ));
    if (echo_12 == 2) {
      ImageP_2 = (float ***) callock( Total - First + 1, sizeof( void * ));
    }
  }
  else {
    SliceNo = Last - First + 1;
    ImageP_1 = alloc_3Dimg( N, M, SliceNo );
    if (echo_12 == 2) {
      ImageP_2 = alloc_3Dimg( N, M, SliceNo );
    }
  }


/*--------------------------------------------------------------------------*/
/* Read (LastSlice - FirstSlice + 1) slices into a 3D array                 */
/*--------------------------------------------------------------------------*/

  LastMarker = (SliceNo ? SliceNo : Total - First + 1);
  for (k = 0; k < LastMarker; k++) {
    if (!Last) {
      ImageP_1[k] = alloc_img( N, M );
    }
    read_matrix(File_1, ImageP_1[k], N, M, ImgType1);
    if (!Last) {
      for (i = N * M - 1; i >= 0; i--)
	if (*(ImageP_1[k][0] + i))
	  break;
      if (i < 0) {
	/* Oops, we have hit the first empty slice; rewind to its start */
	fseek( File_1, -(N * M * abs(ImgType1)), SEEK_CUR );
	free( ImageP_1[k] );
	ImageP_1[k] = NULL;
	Last = First + k - 1;
	SliceNo = Last - First + 1; /* Yeah, it's k, but this is the real
				       meaning behind SliceNo. */
	k = LastMarker;
	printf( "Last slice: %d\n", Last);
      }
      else {
	if (echo_12 == 2) {
	  ImageP_2[k] = alloc_img( N, M );
	  read_matrix(File_2, ImageP_2[k], N, M, ImgType2);
	}
      }
    }
    else if (echo_12 == 2) {
      read_matrix( File_2, ImageP_2[k], N, M, ImgType2 );
    }
  }

  /*  If all slices are being used, alert user, and make sure Last is
   *  set */
  if (!Last) {
    fprintf( stderr,
	     "WARNING: No empty slices before end of image.  Setting\n" );
    fprintf( stderr,
	     "last slice with brain parenchyma to actual last slice of\n" );
    fprintf( stderr,
	     "image, %d.\n", Total );
    Last = Total;
    SliceNo = Last - First + 1;
  }

  /* Initialize LabelP to binary mask of first echo image */
  LabelP = alloc_3Dimg( N, M, SliceNo );
  for (k = 0; k < SliceNo; k++) {
    ptr_1 = ImageP_1[k][0];
    ptr_2 = LabelP[k][0];
    for (i = N * M; i > 0; i--) {
      *ptr_2++ = (*ptr_1++ == BACKGROUND ? BACKGROUND : OBJECT);
    }
  }


/*--------------------------------------------------------------------------
 *  Segment image(s) using K-mean method: step 1
 *--------------------------------------------------------------------------*/

  SegImgP = alloc_3Dchar( N, M, SliceNo );
  Kmean_1.K_Mean = (float *) callock(NoRegion, sizeof(float));
  Kmean_1.Kmean_Sum = (double *) callock(NoRegion, sizeof(double));
  Kmean_1.Kmean_Count = (int *) callock(NoRegion, sizeof(int));
  if (echo_12 == 2) {
    Kmean_2.K_Mean = (float *) callock(NoRegion, sizeof(float));
    Kmean_2.Kmean_Sum = (double *) callock(NoRegion, sizeof(double));
    Kmean_2.Kmean_Count = (int *) callock(NoRegion, sizeof(int));
  }
  NoVoxel = (int *) callock( NoRegion, sizeof(int) );


  if ( echo_12 == 1 ) {
    Kmean3D_1echo(ImageP_1,
		  LabelP, SegImgP, N, M, SliceNo, NoRegion, 
		  Perturb, Epsilon, GibbsPotential,
		  Weight_1, Kmean_1, 
		  Gibbs3D);
  }
  else if ( echo_12 == 2) {
    Kmean3D_2echo(ImageP_1, ImageP_2,
		  LabelP, SegImgP, N, M, SliceNo, NoRegion,
		  Perturb, Epsilon, GibbsPotential, 
		  Kmean_1, Kmean_2, Weight_1, Weight_2,
		  Gibbs3D);
  }

  for ( k=0; k<NoRegion; k++ ) {
    printf("Mean[%d] = %f \n", k+1, Kmean_1.K_Mean[k]);
  }

  for ( k=0; k<NoRegion; k++ ) {
    NoVoxel[k] = 0;
  }
  Intensity_v = 250.0/NoRegion;
  for ( k=0; k<SliceNo; k++ ) {

    ptr_1 = LabelP[0][0];
    ptr_2 = TmpImgP[0]; 
    ptr_c = SegImgP[k][0];

    for ( i=0;  i<N*M; i++ ) {
      ir = *ptr_c;
      if ( ir != BACKGROUND ) {
	*ptr_1 = Kmean_1.K_Mean[ir - 1];
	NoVoxel[ir-1] ++;
	*ptr_2 = ir*Intensity_v;
      }
      else {
	*ptr_1 = 0.0;
	*ptr_2 = 0.0;
      }
      ptr_1++;
      ptr_2++;
      ptr_c++;
    }
    write_matrix (File_Out, TmpImgP, N, M, UNSIGNED_CHAR);
    if (savemean) {
      write_matrix (File_Mean, LabelP[0], N, M, ImgType1);
    }
  }

  for ( k=0; k<NoRegion; k++ ) {
    printf("Voxel No. of Tissue Type %d = %d \n", k+1, NoVoxel[k]);
  }


/*--------------------------------------------------------------------------
 *  Append image or segmentation slices from LastSlice+1 to the end
 *--------------------------------------------------------------------------*/

  if (savemean) {
    for (k = Last+1; k <= Total; k++) {
      fread( (void *) TmpImgP[0], abs(ImgType1), N * M, File_1 );
      fwrite( (const void *) TmpImgP[0], abs(ImgType1), N * M, File_Mean );
    }
  }
  memset( (void *) TmpImgP[0], 0, N * M );
  for (k = Last + 1; k <= Total; k++) {
    fwrite( (const void *) TmpImgP[0], N * M, 1, File_Out );
  }


/*--------------------------------------------------------------------------*/
/*Close opened files and release memory                                     */
/*--------------------------------------------------------------------------*/

  free(NoVoxel);
  free_img(TmpImgP);
  free_3Dimg(LabelP, SliceNo);
  free_3Dchar(SegImgP, SliceNo);

  free(Kmean_1.K_Mean);
  free(Kmean_1.Kmean_Sum);
  free(Kmean_1.Kmean_Count);
  free_3Dimg(ImageP_1, SliceNo);

  if ( echo_12 == 2) {
    fclose(File_2);
    free(Kmean_2.K_Mean);
    free(Kmean_2.Kmean_Sum);
    free(Kmean_2.Kmean_Count);
    free_3Dimg(ImageP_2, SliceNo);
  }


close_one:
  fclose(File_1);

close_mean:
  if (savemean) {
    fclose(File_Mean);
  }

close_out:
  fclose(File_Out);

close_none:
  free (ptr);

  return retval;
}
