/*--------------------------------------------------------------------------
Kmean_3Dseg_main.c                                                       

Programmer: Michelle Yan
Revisions:  Paul Hughett, Dan Widyono

Copyright (c) 1998-2000 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


   It segments an image using K-mean (ML or MAP) method. If interested in 
only brain volume, BrainMask should be used prior.                     
       Including:                                                         
          Kmean3D_1echo();                                                
          Kmean3D_2echo();                                                
--------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
Some pre-RCS history:
 *   DRW 19980131: optimization
 *   DRW 19980225: kept interactive code in main(), placed rest in
 *	separate function which can in the future be called by Tcl/Tk
 *      or some other external program.
 *   PWH 19980310: ported to Linux, fixed gcc errors and warnings.
----------------------------------------------------------------------------*/

static const char rcsid[] = "$Id: Kmean_3Dseg_main.c,v 1.14 2000/06/05 17:17:59 hughett Exp $ Release $Name: segm-0-90 $";

/*--------------------------------------------------------------------------*/
/* Include proper constant, variable and function declarations              */
/*--------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>   /* For getopt and friends */

#include <sys/types.h>
#include <sys/stat.h>

#include "config.h"

#include "Kmean_3Dseg.h"


#define MAX_INPUT_LEN 255


/* FIXME:  There is an incompatibility between the Solaris and Gnu
   C library implementations of getsubopt.  Both apparently define 
   this function and do so compatibly, but the Gnu stdlib.c does not
   normally provide a prototype for it.  (The prototype is present but
   protected by some #ifdefs that make other changes that I don't want.)
   The solution is to provide the necessary prototype here but only
   for the Gnu compiler.  This may not fit all circumstances, and you
   might need to tweak this for your own system; if so, you also need
   to fix AdpKmean_3Dseg_man.c the same way.  If getsubopt is absent,
   then the -s and -v options are ignored; fortunately, they are rarely
   used anyway.  -PWH 19990123
   FIXME:  Actually, life is even more complex than that.  For gcc on
   Solaris (installed using the Walnut Creek Toolkit for Solaris, Jan 1999)
   getsubopt *is* declared, but with a slightly different prototype.
   So we now have an exception to the exception.  -PWH 20000124 */

#if defined(HAVE_GETSUBOPT) && __GNUC__ && !defined(__sun)
extern int getsubopt(char **optionp, char **tokens, char **valuep);
#endif

#ifdef HAVE_GETSUBOPT     /* FIXME: Needed only if getsubopt is supported */
static char * sliceoptions[] = {
#define START_SLICE 0
  "start",
#define END_SLICE 1
  "end",
  NULL
};

static char * voxelsizeoptions[] = {
#define RES_XY 0
  "xy",
#define RES_Z 1
  "z",
  NULL
};
#endif


static void
showusage( char * argv0 )
{
  fprintf( stderr, "Usage:\n\n" );

  fprintf( stderr, "Batch mode:\n" );
  fprintf( stderr, "	%s [-r rows] [-c cols] [-d] [-m]\n", argv0 );
  fprintf( stderr, "		[-p perturbation] [-h distortion_threshold]\n" );
  fprintf( stderr, "		[-g gibbs_potential] [-t type]\n" );
  fprintf( stderr, "		[-s slice_options] [-v voxel_size_options]\n" );
  fprintf( stderr, "		filename_echo1 standard_deviation_echo1\n" );
  fprintf( stderr, "		[filename_echo2 standard_deviation_echo2]\n\n" );

  fprintf( stderr, "Interactive / Parameter file mode (old method):\n" );
  fprintf( stderr, "	%s < {parameter_filename}\n\n", argv0 );

  fprintf( stderr, "	%s\n\n", argv0 );

  fprintf( stderr, "Batch mode options:\n" );
  fprintf( stderr, "	-r, -c	# of rows and columns; default is 256\n" );
  fprintf( stderr, "	-d	debug\n" );
  fprintf( stderr, "	-m	save mean value image for analysis\n" );
  fprintf( stderr, "	-n	number of segmentation bins; default is 5\n" );
  fprintf( stderr, "	-t	image data type (see table below)\n" );
  fprintf( stderr, "		default is 2\n" );
  fprintf( stderr, "	-p	perturbation; default is 0.01\n" );
  fprintf( stderr, "	-h	distortion threshold; default is 0.02\n" );
  fprintf( stderr, "	-g	Gibbs potential; default is 0\n" );
  fprintf( stderr, "	-s	specify slice options in a comma-separated\n" );
  fprintf( stderr, "		list with no intervening spaces.  The\n" );
  fprintf( stderr, "		following options are supported:\n\n" );

  fprintf( stderr, "		start=slice	where slice is the number\n" );
  fprintf( stderr, "				of the first slice which\n" );
  fprintf( stderr, "				contains brain matter\n\n" );

  fprintf( stderr, "		end=slice	where slice is the number\n" );
  fprintf( stderr, "				of the last slice which\n" );
  fprintf( stderr, "				contains brain matter\n\n" );

  fprintf( stderr, "		Slice numbers begin with 1.  If either\n" );
  fprintf( stderr, "		suboption is not provided, then the first\n" );
  fprintf( stderr, "		and/or last non-empty slices (any voxel has a\n" );
  fprintf( stderr, "		value greater than zero) will be used.\n" );
  fprintf( stderr, "		If a suboption is provided without a value\n" );
  fprintf( stderr, "		(e.g. \"-s start=13,end\" ) then the\n" );
  fprintf( stderr, "		first and/or last slice will be used\n" );
  fprintf( stderr, "		(either 1 or the calculated last slice #).\n" );
  fprintf( stderr, "	-v	specify voxel dimension options in a comma-\n" );
  fprintf( stderr, "		separated list with no intervening spaces.\n" );
  fprintf( stderr, "		The following options are supported:\n\n" );

  fprintf( stderr, "		xy=dimension	size (in mm) of x and y sides of\n" );
  fprintf( stderr, "				(square) pixel; default is 0.9375mm\n" );

  fprintf( stderr, "		z=dimension	size (in mm) of z side of pixel;\n" );
  fprintf( stderr, "				default is 1.0mm\n\n" );

  fprintf( stderr, "Image types:\n");
  fprintf( stderr, "	 1 = One Byte Per Voxel,   Unsigned\n");
  fprintf( stderr, "	 2 = Two Bytes Per Voxel,  Unsigned\n");
  fprintf( stderr, "	-2 = Two Bytes Per Voxel,  Signed\n");
  fprintf( stderr, "	 4 = Four Bytes Per Voxel, Signed\n");
  fprintf( stderr, "	-4 = Floating Point Voxel Data\n");
  fprintf( stderr, "(Usually 2 is correct for raw MRI data)\n");
}


/*--------------------------------------------------------------------------*/
/*  Reading the specifications of the input image files.                    */
/*--------------------------------------------------------------------------*/

/*
 * For MHCRC subjects:
 *
 * 1mm:
 *
 * 1 5
 * 0.01 0.02 0
 * mri_ns.img 2 6
 * 256 256 <STARTSLC> <ENDSLC> <TOTALSLC>
 * 0.9375 1
 *
 * Where mri_ns is the T1-weighted, brainmasked image.
 *
 *
 * 5mm:
 *
 * 2 5
 * 0.01 0.02 0
 * mri_t2_ns.img 2 9
 * mri_pr.img 2 11.25
 * 256 256 <STARTSLC> <ENDSLC> <TOTALSLC>
 * 0.8594 5
 *
 * Where mri_t2_ns is the T2-weighted, brainmasked image and
 *  mri_pr is the Proton-Density image (does not need to be brainmasked
 *  since mask is created from T2-weighted image).  0.8594 is mostly
 *  true; some very old images were 0.7812 mm per voxel in-plane width.
 */

int
main( int argc, char *argv[] )
{
  int		savemean = 0;
  int		echo_12, NoRegion;
  int		ImgType1, ImgType2;
  int		N, M, First, Last, Total;
  char          Name1[128], Name2[128];
  float         Perturb, Epsilon, GibbsPotential; 
  float         Weight_1, Weight_2;
  float         Res_XY, Res_Z;
  struct stat	filestat;

  int		c;
  int		debugflag = 0;

#ifdef HAVE_GETSUBOPT           /* Needed only if getsubopt is supported */
  char		*options, *value;
#endif

  /* Default settings */
  ImgType1 = ImgType2 = 2;	/* unsigned 2-byte */
  N = M = 256;			/* 256 x 256 pixels per slice */
  First = Last = Total = 0;	/* calculate # slices using file size */
  Perturb = 0.01;
  Epsilon = 0.02;
  GibbsPotential = 0;
  savemean = 0;
  Res_XY = 0.9375;
  Res_Z = 1.0;
  echo_12 = 1;
  NoRegion = 5;


  while ((c = getopt( argc, argv, "dmn:r:c:t:p:h:g:s:v:")) != EOF) {
    switch (c) {

    case 'd':
      debugflag = 1;
      break;

    case 'm':
      savemean = 1;
      break;

    case 'n':
      if (sscanf( optarg, "%d", &NoRegion ) != 1) {
	fprintf( stderr, "Error parsing number of regions.  Using 5.\n" );
	NoRegion = 5;
      }
      break;

    case 'r':
      if (sscanf( optarg, "%d", &N ) != 1) {
	fprintf( stderr, "Error parsing number of rows.  Using 256.\n" );
	N = 256;
      }
      break;

    case 'c':
      if (sscanf( optarg, "%d", &M ) != 1) {
	fprintf( stderr, "Error parsing number of columns.  Using 256.\n" );
	M = 256;
      }
      break;

    case 't':
      /* For now, both images must be of the same type!  */
      if (sscanf( optarg, "%d", &ImgType1 ) != 1) {
	fprintf( stderr, "Error parsing image type.  Using 2 (unsigned 2-byte).\n" );
	ImgType1 = ImgType2 = 2;
      }
      ImgType2 = ImgType1;
      break;

    case 'p':
      if (sscanf( optarg, "%e", &Perturb) != 1) {
	fprintf( stderr, "Error parsing perturbation.  Using 0.01.\n" );
	Perturb = 0.01;
      }
      break;

    case 'h':
      if (sscanf( optarg, "%e", &Epsilon) != 1) {
	fprintf( stderr, "Error parsing distortion threshold.  Using 0.02.\n" );
	Epsilon = 0.02;
      }
      break;

    case 'g':
      if (sscanf( optarg, "%e", &GibbsPotential) != 1) {
	fprintf( stderr, "Error parsing gibbs potential.  Using 0.\n" );
	GibbsPotential = 0;
      }
      break;

      /* FIXME:  The -s and -v options are supported only if getsubopt
         is available, which it isn't under Linux. */
#ifdef HAVE_GETSUBOPT
    case 's':
      options = optarg;
      while (*options != '\0') {
	switch (getsubopt( &options, sliceoptions, &value )) {

	case START_SLICE:
	  if (value == NULL)
	    First = 1;
	  else if (sscanf( value, "%d", &First) != 1) {
	    fprintf( stderr, "Error parsing start slice number.  Will search for\n" );
	    fprintf( stderr, "first non-empty slice instead.\n" );
	    First = 0;
	  }
	  break;

	case END_SLICE:
	  if (value == NULL)
	    Last = -1;
	  else if (sscanf( value, "%d", &Last) != 1) {
	    fprintf( stderr, "Error parsing end slice number.  Will search for\n" );
	    fprintf( stderr, "last non-empty slice instead.\n" );
	    Last = 0;
	  }
	  break;

	default:
	  showusage( argv[0] );
	  exit( 1 );
	}
      }
      break;
#endif

#ifdef HAVE_GETSUBOPT     /* FIXME:  Klutzy patch to compile under Linux */
    case 'v':
      options = optarg;
      while (*options != '\0') {
	switch (getsubopt( &options, voxelsizeoptions, &value )) {

	case RES_XY:
	  if (value == NULL) {
	    showusage( argv[0] );
	    exit( 1 );
	  }
	  if (sscanf( value, "%e", &Res_XY) != 1) {
	    fprintf( stderr, "Error parsing xy voxel dimension.  Using 0.9375mm.\n" );
	    Res_XY = 0.9375;
	  }
	  break;

	case RES_Z:
	  if (value == NULL) {
	    showusage( argv[0] );
	    exit( 1 );
	  }
	  if (sscanf( value, "%e", &Res_Z) != 1) {
	    fprintf( stderr, "Error parsing z voxel dimension.  Using 1.0mm.\n" );
	    Res_Z = 1.0;
	  }
	  break;

	default:
	  showusage( argv[0] );
	  exit( 1 );
	}
      }
      break;
#endif

    case '?':
      showusage( argv[0] );
      exit( 1 );
    }
  }


  if (argc == 1 || ((argc == 2) && (debugflag))) {

    /*
     * No filename(s) were given, and no options were given.
     * Revert to old method of getting parameters from stdin.
     */

    /* silence stdout for batch mode (e.g. "Kmean_3Dseg < parameter_file") */
    if (!(isatty( fileno( stdin )))) {
      if (!debugflag)
	freopen( "/dev/null", "w", stdout );
    }
    else {
      showusage( argv[0] );
    }

    printf("Number of echoes (1/2), number of regions:\n");
    scanf("%d %d", &echo_12, &NoRegion);
    printf("Number of echoes, number of regions = %d, %d\n", 
	   echo_12, NoRegion);

    printf("Perturbation, distortion threshold & Gibbs potential:\n");
    scanf("%e %e %e", &Perturb, &Epsilon, &GibbsPotential);
    printf("Perturbation, distortion & potential = %0.3e, %0.3e, %0.3e\n",
	   Perturb, Epsilon, GibbsPotential);

    printf("** Set 1 should already have been brainmasked. **\n");
    printf("Image name, type & standard deviation (set 1 of %d): \n",
	   echo_12);
    scanf("%s %d %f", Name1, &ImgType1, &Weight_1);
    printf("Image name, type & standard deviation: %s (%d), %0.3f\n",
	   Name1, ImgType1, Weight_1);
    Weight_1 = 0.5 / ( Weight_1*Weight_1);

    if ( echo_12 == 2 ) {
      printf("Image name, type & standard deviation of set 2: \n");
      scanf("%s %d %f", Name2, &ImgType2, &Weight_2);
      printf("Image name, type & standard deviation: %s (%d), %0.3f\n",
	     Name2, ImgType2, Weight_2);
      Weight_2 = 0.5 / (Weight_2 * Weight_2);
    }

    printf("Image dimensions:  Slice numbers start from 1.\n");
    printf("(Rows Cols First_slice Last_slice Total_slices): \n");
    scanf("%d %d %d %d %d", &N, &M, &First, &Last, &Total);
    printf("Image dimension = %d X %d; %d -> %d  (%d)\n", 
	   N, M, First, Last, Total);

    printf("Voxel size (XY- & Z-resolution):\n");
    scanf("%e %e", &Res_XY, &Res_Z);
    printf("XY- & Z-resolution = %0.3e %0.3e\n", Res_XY, Res_Z);


    return kmean_3dseg( echo_12, NoRegion,
			Name1, ImgType1, Weight_1,
			Name2, ImgType2, Weight_2,
			N, M, First, Last, Total,
			Res_XY, Res_Z,
			Perturb, Epsilon, GibbsPotential,
			savemean );
  }


  /* silence stdout during normal operation */
  if (!debugflag)
    freopen( "/dev/null", "w", stdout );

  if (optind == argc) {
    showusage( argv[0] );
    fprintf( stderr, "Please provide a filename and standard deviation.\n" );
    exit( 1 );
  }

  if ((argc - optind) % 2) {
    showusage( argv[0] );
    fprintf( stderr, "Each filename must be matched with a standard deviation.\n" );
    exit( 1 );
  }

  if ((argc - optind) > 4) {
    showusage( argv[0] );
    fprintf( stderr, "Maximum of two echoes allowed.\n" );
    exit( 1 );
  }

  if ((argc - optind) == 4)
    echo_12 = 2;

  printf("optind = %d, argc = %d\n", optind, argc);
  {
    int i;
    for (i = 0; i < argc; i++)
      printf("argv[%d] = \"%s\"\n", i, argv[i]);
  }

  strncpy( Name1, argv[optind++], 126 );
  if (stat( Name1, &filestat ) < 0) {
    perror( Name1 );
    putchar( '\n' );
    exit( -1 );
  }
  Total = filestat.st_size / (N * M * abs(ImgType1));
  if (Last == -1)
    Last = Total;
  if (sscanf( argv[optind++], "%e", &Weight_1 ) != 1) {
    showusage( argv[0] );
    fprintf( stderr, "Error parsing standard deviation for echo 1.\n" );
    exit( 1 );
  }

  if (echo_12 == 2) {
    strncpy( Name2, argv[optind++], 126 );
    if (sscanf( argv[optind], "%e", &Weight_2 ) != 2) {
      showusage( argv[0] );
      fprintf( stderr, "Error parsing standard deviation for echo 2.\n" );
      exit( 1 );
    }
  }



  printf("Number of echoes, number of regions = %d, %d\n", 
	 echo_12, NoRegion);
  printf("Perturbation, distortion & potential = %0.3e, %0.3e, %0.3e\n",
	 Perturb, Epsilon, GibbsPotential);
  printf("Image name, type & standard deviation: %s (%d), %0.3f\n",
	 Name1, ImgType1, Weight_1);
  Weight_1 = 0.5 / ( Weight_1*Weight_1);
  if ( echo_12 == 2 ) {
    printf("Image name, type & standard deviation: %s (%d), %0.3f\n",
	   Name2, ImgType2, Weight_2);
    Weight_2 = 0.5 / (Weight_2 * Weight_2);
  }
  printf("Image dimension = %d X %d; %d -> %d  (%d)\n", 
	 N, M, First, Last, Total);
  printf("XY- & Z-resolution = %0.3e %0.3e\n", Res_XY, Res_Z);


  return kmean_3dseg( echo_12, NoRegion,
		      Name1, ImgType1, Weight_1,
		      Name2, ImgType2, Weight_2,
		      N, M, First, Last, Total,
		      Res_XY, Res_Z,
		      Perturb, Epsilon, GibbsPotential,
		      savemean );
}
