/***************************************************************************
AdpKmean_3Dseg_main.c - Adaptive K-means segmentation

Programmer: Paul Hughett

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This program segments an MRI volume image of the brain into CSF, gray
matter, and white matter using an adaptive K-means algorithm developed
by Michelle Yan.  It expects two or three files as input, and writes
one or two files as output.  The input files are one or two raw image
files containing one or two echoes of the image to be segmented; and
an initial segmentation (usually done by Kmean_3Dseg), contained in a
file whose name is constructed by prefixing "ml_" to the name of the
file containing the first echo.  The output files are the segmented
image, contained in a file whose name is constructed by prefixing
"Adpml_" to the filename of the first echo, and an optional means
files, which always has the name "Adpmean.img".

This program contains several tuning parameters which match the
algorithm to the characteristics of particular classes of images.
These are determined by experiment and the appropriate set of values
is selected according to such accidental values as the ratio between
the in-slice and between-slice spacings.  This really ought to be done
in some more rational and portable fashion.

****************************************************************************/

static const char rcsid[] = "$Id: adpkm.c,v 1.3 2000/06/05 17:17:59 hughett Exp $ Release $Name: segm-0-90 $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <time.h>
#include <unistd.h>       /* For getopt function */

#include <sys/types.h>
#include <sys/stat.h>

#include "config.h"

#include "errm.h"
#include "AdpKmean_3D.h"
#include "param.h"


int
main( int argc, char *argv[] )
{
    Param_list params;             /* Contents of param file */
    char *stage = "adpkm";         /* Default stage name for params */
    const char *image1, *image2;   /* Filenames for the two echo images */
    int	savemean = 0;              /* Save the means image? */
    int	echo_12;                   /* Number of echoes: 1 or 2 */
    int NoRegion;                  /* Number of regions to classify into */
    int	ImgType1, ImgType2;        /* Types of 1st and 2nd images */
    int nrows, ncols;              /* Number of rows and columns */
    int nslices;                   /* Number of slices */
    int First, Last;               /* First and last slice to process */
    float Res_XY, Res_Z;           /* Physical voxel dimensions [mm] */
    float std1;                    /* Standard deviation of noise in image 1 */
    float std2;                    /* Standard deviation of noise in image 2 */
    float weight1, weight2;        /* Relative weights for images 1 & 2 */
    int flag;                      /* Flag returned by segmentation engine */
    struct stat	filestat;
    const char *s;
    clock_t clock0, clock1;        /* To compute CPU run time */
    time_t  time0,  time1;         /* To compute elapsed time */
    double  cpu_time, elapsed;

    /* Record start times */
    clock0 = clock();
    time0  = time(NULL);

    /* Read the param file named on the command line */
    if (argc < 2)
	fatal("No param file specified");
    params = read_param_file(argv[1]);

    /* Extract the necessary parameter values */
    image1 = get_param(params, stage, "image1");
    image2 = get_param(params, stage, "image2");
    ImgType1 = (s = get_param(params, stage, "type1"   ) ) ? atoi(s) :   2;
    ImgType2 = (s = get_param(params, stage, "type2"   ) ) ? atoi(s) :   2;
    nrows    = (s = get_param(params, stage, "nrows"   ) ) ? atoi(s) : 256;
    ncols    = (s = get_param(params, stage, "ncols"   ) ) ? atoi(s) : 256;
    nslices  = (s = get_param(params, stage, "nslices" ) ) ? atoi(s) :   0;
    First    = (s = get_param(params, stage, "first"   ) ) ? atoi(s) :   1;
    Last     = (s = get_param(params, stage, "last"    ) ) ? atoi(s) :  -1;
    echo_12  = (s = get_param(params, stage, "nechoes" ) ) ? atoi(s) :   1;
    NoRegion = (s = get_param(params, stage, "nregions") ) ? atoi(s) :   5;
    std1     = (s = get_param(params, stage, "std1"    ) ) ? atof(s) : 10.0;
    std2     = (s = get_param(params, stage, "std2"    ) ) ? atof(s) : 10.0;
    weight1  = (s = get_param(params, stage, "weight1" ) ) ? atof(s) : 0.12;
    weight2  = (s = get_param(params, stage, "weight2" ) ) ? atof(s) : 0.44;
    Res_XY   = (s = get_param(params, stage, "xspace"  ) ) ? atof(s) : 0.9375;
    Res_XY   = (s = get_param(params, stage, "yspace"  ) ) ? atof(s) : Res_XY;
    Res_XY   = (s = get_param(params, stage, "xyspace" ) ) ? atof(s) : Res_XY;
    Res_Z    = (s = get_param(params, stage, "zspace"  ) ) ? atof(s) : 1.0;

    /* FIXME:  Provide reasonable defaults for input and output file names */

    /* FIXME: Add consistency checks for the parameters */

    /* FIXME: The reason for the following code is rather obscure, 
       so just comment out for now rather than delete. -PWH 19990121 */
#if 0
    /* silence stdout during normal operation */
    if (!debugflag)
	freopen( "/dev/null", "w", stdout );
#endif

    /* Determine nslices from size of image file */
    if (nslices <= 0) {
	if (stat( image1, &filestat ) < 0) 
	    fatal ("Unable to read size of image1 %s", image1);
	nslices = filestat.st_size / (nrows * ncols * abs(ImgType1));
    }

    /* Last defaults to the number of slices */
    if (Last == -1)
	Last = nslices;

    /* Print out the actual parameter values for inspection */
    printf("Number of echoes, number of regions = %d, %d\n", 
	   echo_12, NoRegion);
    if (echo_12 == 1) {
	printf("Image name, type & standard deviation: %s (%d), %0.3f\n",
	       image1, ImgType1, std1);
	std1 = 0.5 / (std1*std1);   /* FIXME:  Hunh?  What's this? */
    }
    else if ( echo_12 == 2 ) {
	printf("Image name, type, STD & weight: %s (%d), %0.3f, %0.3f\n",
	       image1, ImgType1, std1, weight1);
	printf("Image name, type, STD & weight: %s (%d), %0.3f, %0.3f\n",
	       image2, ImgType2, std2, weight2);
    }
    printf("Image dimension = %d X %d; %d -> %d  (%d)\n", 
	   nrows, ncols, First, Last, nslices);
    printf("XY- & Z-resolution = %0.3e %0.3e\n", Res_XY, Res_Z);
    
    /* Now call the segmentation engine to do all the real work */
    flag = adpkmean_3dseg( echo_12, NoRegion,
			   image1, ImgType1, std1, weight1,
			   image2, ImgType2, std2, weight2,
			   nrows, ncols, First, Last, nslices,
			   Res_XY, Res_Z,
			   savemean );

    /* Report the CPU and elapsed time used */
    clock1 = clock();
    time1  = time(NULL);
    cpu_time = (clock1 - clock0) / (double)CLOCKS_PER_SEC;
    elapsed  = (double)(time1 - time0);
    fprintf(stderr, "CPU run time was %8.3f seconds.\n", cpu_time);
    fprintf(stderr, "Elapsed time was %8.3f seconds.\n", elapsed);

    return flag;
}
