/****************************************************************************
dconv.c - Data conversion between internal and external formats

Author:  Paul Hughett

Copyright (c) 1998-2000 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file contains various functions used within bblimage to convert
between internal and external data formats, or between external data
formats used on different platforms.  They are probably useful for
applications other than bblimage as well.

FIXME: The functions in dconv.c were extracted from image_io.c and
represent an older approach to conversion than does exim.c; the two
files dconv.c and exim.c should be inspected for redundant functionality.
*****************************************************************************/

static const char rcsid[] = "$Id: dconv.c,v 1.2 2000/06/05 17:49:33 hughett Exp $ Release $Name: segm-0-90 $";

#include <math.h>
#include <stdlib.h>

#include "errm.h"
#include "memm.h"
#include "dconv.h"



/*------------------------------------------------------------------
Routines to convert from external to internal representations.

The external representations are used for data in binary files
and should be independent of the machine architecture for portability.
All the routines given here use big endian order for the external data.

FIXME:  Add warnings for values that cannot be converted.  Permit
the external format to be little or middle endian.  Add the missing
conversions, including IEEE format.
-------------------------------------------------------------------*/

/* Convert unsigned 1-byte integer format to float */
void
uchar_to_float(          
   long count,            /* Number of elements to process */
   float *dest,           /* Buffer for results */
   unsigned char *src)    /* Buffer of data to be converted */
{
   long i;
   for (i = 0; i < count; i++)
      dest[i] = (float)src[i];
   return;
}

/* Convert unsigned 2-byte integer big endian format to float */
void
uint2_to_float(          
   long count,            /* Number of elements to process */
   float *dest,           /* Buffer for results */
   unsigned char *src)    /* Buffer of data to be converted */
{
   long i;
   for (i = 0; i < count; i++)
      dest[i] = 256.0*(float)src[2*i] + (float)src[2*i+1];
   return;
}

/* Convert 2-byte two's comp big endian integer format to float */
void
int2_to_float(          
   long count,            /* Number of elements to process */
   float *dest,           /* Buffer for results */
   unsigned char *src)    /* Buffer of data to be converted */
{
   long i;
   for (i = 0; i < count; i++)
      if (src[2*i] < 0x80)
         dest[i] = 256.0*(float)src[2*i] + (float)src[2*i+1];
      else
         dest[i] = 256.0*(float)src[2*i] + (float)src[2*i+1] - 32768.0;
   return;
}

/* Convert 4-byte two's comp big endian integer format to float */
/* FIXME:  Provided for compatibility only; not all 4-byte ints will
   fit into a 4-byte float without loss of information; and not all
   that will fit will be correctly translated by this code */
void
int4_to_float(          
   long count,            /* Number of elements to process */
   float *dest,           /* Buffer for results */
   unsigned char *src)    /* Buffer of data to be converted */
{
   long i;
   for (i = 0; i < count; i++)
      if (src[4*i] < 0x80)
         dest[i] = (double)src[4*i+3] + 256.0*((double)src[4*i+2] 
                 + 256.0*((double)src[4*i+1] + 256.0*(double)src[4*i]));
      else
         dest[i] = (double)src[4*i+3] + 256.0*((double)src[4*i+2] 
                 + 256.0*((double)src[4*i+1] + 256.0*(double)src[4*i]))
                 - 2147483648.0;

   return;
}

/* Convert 4-byte two's comp big endian integer format to double */
void
int4_to_double(          
   long count,            /* Number of elements to process */
   double *dest,          /* Buffer for results */
   unsigned char *src)    /* Buffer of data to be converted */
{
   long i;
   for (i = 0; i < count; i++)
      if (src[4*i] < 0x80)
         dest[i] = (double)src[4*i+3] + 256.0*((double)src[4*i+2] 
                 + 256.0*((double)src[4*i+1] + 256.0*(double)src[4*i]));
      else
         dest[i] = (double)src[4*i+3] + 256.0*((double)src[4*i+2] 
                 + 256.0*((double)src[4*i+1] + 256.0*(double)src[4*i]))
                 - 2147483648.0;
   return;
}


/*-----------------------------------------------------------------
swap_bytes - Change byte order

Given a byte string containing count blocks of nbin bytes each, this
function writes out those blocks with nbout bytes each by rearranging
the input bytes and possibly inserting zero bytes or omitting input
bytes.  Note that this is a generic function for an arbitrary new byte
order; in many common cases, there is a specific byte-swapping routine
that will be faster.  The order of the bytes in the output is
specified by an integer array inbyte in which the ith entry specifies
the source of the ith output byte in each block.  If the ith entry in
this array is zero or positive, then it is the number of the input
byte which should be used as the ith output byte; if it is negative,
then the value of ctab[1-inbyte[i]] is used for the ith output byte.
By special dispensation, supplying a null pointer for ctab will set
the output byte to zero.

The results are unpredictable if the source and destination buffers
are the same.

TECHNICAL NOTES:  We assume that the compiler is smart enough to do
strength reduction and the other simple loop optimizations and write
the loops to be understood by a human.  For efficiency, we loop over
the output byte numbers in the outer loop, and over blocks in the
inner loop.  The inner loop should not be too long, or the cache will
miss and slow things down.  

FIXME: Block the outer loop to approximate the cache size so that the
caller doesn't have to.

-------------------------------------------------------------------*/

void
swap_bytes(
    long count,            /* Number of blocks to process */
    unsigned char *dest,   /* Destination buffer */   
    unsigned char *src,    /* Source buffer */
    int nbout,             /* Number of bytes per output block */
    int nbin,              /* Number of bytes per input block */
    int *inbyte,           /* Input byte number for each output byte */
    unsigned char *ctab)   /* Table of constant values */
{
    int i;                 /* Index over output bytes per block */
    long k;                /* Index over output blocks */
    unsigned char c1;
    int c2;

    for (i = 0; i < nbout; i++) {
	if (inbyte[i] < 0) {
	    c1 = (ctab == 0) ? 0 : ctab[1-inbyte[i]];
	    for (k = 0; k < count; k++)
		dest[k*nbout+i] = c1;
	}
	else {
	    c2 = inbyte[i];
	    for (k = 0; k < count; k++)
		dest[k*nbout+i] = src[k*nbin+c2];
	}
    }

    return;
}

/*------------------------------------------------------------------
Routines to convert from internal to external representations.

These routines have been written rather sloppily and do not promise
to produce the right results for image data outside a reasonable
range; by definition, it's unreasonable if these routines munge it.
In any case, the float data are rounded to the nearest integer, or
set to the appropriate limiting value.

Some of these routines permit the endianness and number of bytes per
datum to be specified as an argument.  In this case, the term datum
refers to a single item in the internal representation (float, int,
long, char, etc), which is converted to a variable number of bytes
(almost always unsigned char) arranged in the specified byte order.
The count always refers to the number of data in the internal format.
If the number of bytes specified is less than the number actually
present in the datum, the high order bytes are omitted.  For efficiency,
these routines typically dispatch on the number of bytes and byte
order and have a custom loop for each possible combination (but it
is assumed that the compiler will optimize the details).
-------------------------------------------------------------------*/

void
ulong_to_bytes(
    long count,           /* Number of data to process */
    unsigned char *dest,  /* Output buffer */
    unsigned long *src,   /* Input buffer */
    int nbytes,           /* Number of bytes per datum: 1, 2, 3, or 4 */
    int bigendian)        /* Big (1) or little (0) endian */
{
    long i;

    /* NOTE: This code exploits the fact that, in ANSI C, assigning an
       unsigned long to an unsigned char simply drops the high-order
       bits to avoid explicitly masking with 0xff. If your compiler is
       non conformant in this respect, you may want to add the
       explicit mask. */

    if (nbytes < 1 || nbytes > 4) {
	fatal("Invalid nbytes in ulong_to_uchar"); }
 
    else if (nbytes == 4 && bigendian) {
	for ( i = 0; i < count; i++) {
	    dest[4*i+3] = src[i];
	    dest[4*i+2] = src[i] >>  8;
	    dest[4*i+1] = src[i] >> 16;
	    dest[4*i  ] = src[i] >> 24;
	}
    }

    else if (nbytes == 4) {
	for ( i = 0; i < count; i++) {
	    dest[4*i  ] = src[i];
	    dest[4*i+1] = src[i] >>  8;
	    dest[4*i+2] = src[i] >> 16;
	    dest[4*i+3] = src[i] >> 24;
	}
    }

    else if (nbytes == 3 && bigendian) {
	for ( i = 0; i < count; i++) {
	    dest[3*i+2] = src[i];
	    dest[3*i+1] = src[i] >>  8;
	    dest[3*i  ] = src[i] >> 16;
	}
    }

    else if (nbytes == 3) {
	for ( i = 0; i < count; i++) {
	    dest[3*i  ] = src[i];
	    dest[3*i+1] = src[i] >>  8;
	    dest[3*i+2] = src[i] >> 16;
	}
    }

    else if (nbytes == 2 && bigendian) {
	for ( i = 0; i < count; i++) {
	    dest[2*i+1] = src[i];
	    dest[2*i  ] = src[i] >>  8;
	}
    }

    else if (nbytes == 2) {
	for ( i = 0; i < count; i++) {
	    dest[2*i  ] = src[i];
	    dest[2*i+1] = src[i] >>  8;
	}
    }

    else if (nbytes == 1) {
	for ( i = 0; i < count; i++) {
	    dest[i] = src[i];
	}
    }

    return;
}


void
float_to_uchar(
   long count,           /* Number of data to process */
   unsigned char *dest,  /* Output buffer */
   float *src)           /* Input buffer */
{
   long i;

   for ( i = 0; i < count; i++) {
      if (src[i] <= 0.5)
         dest[i] = 0;
      else if (src[i] >= 255.0)
         dest[i] = 255; 
      else
         dest[i] = floor(src[i]+0.5);
   }
   return;
}


void
float_to_uint2(
   long count,           /* Number of data to process */
   unsigned char *dest,  /* Output buffer */
   float *src)           /* Input buffer */
{
   long i, val;

   for ( i = 0; i < count; i++) {
      if (src[i] <= 0.5)
         val = 0;
      else if (src[i] >= 65535.0)
         val = 65535; 
      else
         val = floor(src[i]+0.5);
      dest[2*i]   = val >> 8;
      dest[2*i+1] = val & 0xff;
   }
   return;
}


void
float_to_int2(
   long count,           /* Number of data to process */
   unsigned char *dest,  /* Output buffer */
   float *src)           /* Input buffer */
{
   long i, val;

   for ( i = 0; i < count; i++ ) {
      if (src[i] <= -32768.5) {
	 dest[2*i]   = 0x80;
         dest[2*i+1] = 0; }
      else if (src[i] >= 32767.0) {
	 dest[2*i]   = 0xff;
         dest[2*i+1] = 0xff; }
      else if (src[i] < 0.0) {
         val = floor(src[i] + 32768.5);
         dest[2*i]   = 0x80 + (val >> 8);
         dest[2*i+1] = val & 0xff; }
      else {
         val = floor(src[i]+0.5);
         dest[2*i]   = val >> 8;
         dest[2*i+1] = val & 0xff; }
   }
   return;
}


void
float_to_int4(
   long count,           /* Number of data to process */
   unsigned char *dest,  /* Output buffer */
   float *src)           /* Input buffer */
{
   long i, val;

   for ( i = 0; i < count; i++ ) {
      if      (src[i] <= -2147483648.5) {
	 dest[4*i]   = 0x80;
         dest[4*i+1] = 0; 
         dest[4*i+2] = 0;
         dest[4*i+3] = 0; }
      else if (src[i] >=  2147483647.0) {
	 dest[4*i]   = 0xff;
         dest[4*i+1] = 0xff; 
         dest[4*i+2] = 0xff;
         dest[4*i+3] = 0xff; }
      else if (src[i] < 0.0) {
         val = floor(src[i] + 2147483648.5);
         dest[4*i]   = 0x80 + (val >> 24);
         dest[4*i+1] = 0xff & (val >> 16);
         dest[4*i+2] = 0xff & (val >> 8);
         dest[4*i+3] = 0xff & val;  }
      else {
         val = floor(src[i]+0.5);
         dest[4*i]   =         val >> 24;
         dest[4*i+1] = 0xff & (val >> 16);
         dest[4*i+2] = 0xff & (val >> 8 );
         dest[4*i+3] = 0xff & val; }
   }
   return;
}
