/*--------------------------------------------------------------------------
eigen_vector.c - Image covariance and eigenvector calculations           

Programmer: Michelle Yan

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


   This source code contains some basic functions for finding a eigen     
vector matrix of a 3X3 covariance matrix.                                

Including:                                                             
  int Check_InRegion     : check whether a 3D point is inside a given    
                           region.                                       
  void Find_CovMatrix()  : find a 3X3 scaled covariance matrix of 3D     
                           BOUNDARY.                                     
  void Jacobi_Rotate()   : computes elements (Row_r, Col_p) and          
                           (Row_s, Col_q) for r!=p and s!=q after        
                           applying a Jacobi plane rotation. Note that   
                           input array is double precision and           
                           the same elements of the array are changed    
  int Eigen_SymMatrix()  : diagnalize a double precision symmetric matrix
                           and find its double precision eigen values    
                           and vector matrix using cyclic Jacobi trans.  
  void X_2Matrix         : multiple double matrix(NXM) by double         
                           matrix(MXL)                                   
  void T_Matrix          : transpose a double matrix(NXM)                
--------------------------------------------------------------------------*/

static const char rcsid[] = "$Id: eigen_vector.c,v 1.7 2000/06/05 17:17:59 hughett Exp $ Release $Name: segm-0-90 $";

#include <stdio.h>
#include <math.h>
#include <string.h>

#include "definition.h"
#include "eigen_vector.h"



/*--------------------------------------------------------------------------*/
/* purpose:                                                                 */ 
/*    it checks whethr a 3D point is inside a given region.                 */
/*--------------------------------------------------------------------------*/
/*Input and output variable definitions                                     */
/*--------------------------------------------------------------------------*/

int Check_InRegion(
   float xi, float yj, float zk,           /* Point to check */
   float Low_I, float Low_J, float Low_K,  /* Bounding box limits */
   float Up_I, float Up_J, float Up_K)
{
/*--------------------------------------------------------------------------*/
/*Local variable definitions                                                */
/*--------------------------------------------------------------------------*/
    int   InRegion = TRUE;
    
    if ( xi < Low_I || xi > Up_I ||
	 yj < Low_J || yj > Up_J ||
	 zk < Low_K || zk > Up_K )
	InRegion = FALSE;

    return InRegion;
}


/*--------------------------------------------------------------------------*/
/* Find_Covmatrix                                                           */ 
/*    it finds a Covariance matrix of 3D BOUNDARY, which is scaled by it    */
/* voxel resolution.                                                        */
/*--------------------------------------------------------------------------*/

void Find_CovMatrix(
   struct coordinate2 **B_Coordinate3P,
   int *B_SizeP, 
   CovMatrix_Info *InfoP,  
   int FirstSlice, 
   int LastSlice, 
   float Low_I,             /* Lower & Upper bound of the other  */
   float Low_J,             /* set , so to use region intersected*/
   float Low_K,             /* with the other one                */
   float Up_I, 
   float Up_J, 
   float Up_K, 
   double **RotationP, 
   double *TranslationP)
{
/*--------------------------------------------------------------------------*/
/*Local variable definitions                                                */
/*--------------------------------------------------------------------------*/
    int                  InBoundary_Num, InRegion_YN, SliceNo, i, k;
    float                xi, yj, zk;
    float                tmp_i, tmp_j, tmp_k, Sq_k;
    struct coordinate2   *Ptr_1;

/*--------------------------------------------------------------------------*/
/*Initialize Mean, Covariance matrix and total Obj size                     */
/*--------------------------------------------------------------------------*/
    for ( i=0; i<COVMATRIX_L; i++ ) {
	InfoP->Mean[i] = 0.0;
	for ( k=0; k<COVMATRIX_L; k++ ) {
	    InfoP->Cov[i][k] = 0.0;
	}
    }
    InfoP->Total_ObjSize = 0;

/*--------------------------------------------------------------------------*/
/*Process one BOUNDARY at a time, start at FirstSlice BOUNDARY              */
/*--------------------------------------------------------------------------*/
    SliceNo = LastSlice - FirstSlice + 1;
    printf("B_SizeP[0], SliceNo =: %d %d\n", B_SizeP[0], SliceNo);
    for ( k = 0; k < SliceNo; k++) {

/*--------------------------------------------------------------------------*/
/* Cumulate contribution from each BOUNDARY                                 */
/*--------------------------------------------------------------------------*/
	tmp_k = ((k + FirstSlice - 1) - InfoP->ImgCenter[2]) * 
	        InfoP->Z_Resolution;
	Sq_k = tmp_k * tmp_k;

	printf("B_SizeP[%d], B_Coordinate[0](i,j) =: %d %d %d\n", 
	       k+1, B_SizeP[k+1], B_Coordinate3P[k][0].i, 
	       B_Coordinate3P[k][0].j );
	InBoundary_Num = 0;
	for ( i=0, Ptr_1=B_Coordinate3P[k]; i<B_SizeP[k+1]; Ptr_1++, i++) {
	    tmp_i = (Ptr_1->i - InfoP->ImgCenter[0]) * InfoP->XY_Resolution;
	    tmp_j = (Ptr_1->j - InfoP->ImgCenter[1]) * InfoP->XY_Resolution;
/*--------------------------------------------------------------------------*/
/* Rotate and translate a 3D point to coordinate of the other set, check    */
/* if the transformed one is inside of a given region. If inside, then      */
/* count the point contribution                                             */
/*--------------------------------------------------------------------------*/
	    xi = RotationP[0][0] * tmp_i + RotationP[0][1] * tmp_j + 
		 RotationP[0][2] * tmp_k;
	    yj = RotationP[1][0] * tmp_i + RotationP[1][1] * tmp_j + 
		 RotationP[1][2] * tmp_k;
	    zk = RotationP[2][0] * tmp_i + RotationP[2][1] * tmp_j + 
		 RotationP[2][2] * tmp_k;
	    xi += TranslationP[0];
	    yj += TranslationP[1];
	    zk += TranslationP[2];
			   
	    InRegion_YN = Check_InRegion (xi, yj, zk, Low_I, Low_J, Low_K,
					  Up_I, Up_J, Up_K);
	    if ( InRegion_YN == TRUE ) {
		InBoundary_Num ++;
		
		InfoP->Mean[0] += tmp_i;
		InfoP->Mean[1] += tmp_j;
		InfoP->Mean[2] += tmp_k;

		InfoP->Cov[0][0] += tmp_i * tmp_i;
		InfoP->Cov[0][1] += tmp_i * tmp_j;
		InfoP->Cov[0][2] += tmp_i * tmp_k;
		InfoP->Cov[1][1] += tmp_j * tmp_j;
		InfoP->Cov[1][2] += tmp_j * tmp_k;
		InfoP->Cov[2][2] += Sq_k;
	    }
	}
	InfoP->Total_ObjSize += InBoundary_Num;
    }
    printf("Total number of inside BOUNDARY pixels = %d\n", 
	   InfoP->Total_ObjSize);
/*--------------------------------------------------------------------------*/
/*Update each element of structure CovMatrix_info and 3X3 covariance matrix */
/*--------------------------------------------------------------------------*/
    InfoP->Mean[0] /= InfoP->Total_ObjSize;
    InfoP->Mean[1] /= InfoP->Total_ObjSize;
    InfoP->Mean[2] /= InfoP->Total_ObjSize;

    InfoP->Cov[0][0] = (InfoP->Cov[0][0]/InfoP->Total_ObjSize) - 
	               (InfoP->Mean[0]) * (InfoP->Mean[0]);
    InfoP->Cov[0][1] = (InfoP->Cov[0][1]/InfoP->Total_ObjSize) - 
     	               (InfoP->Mean[0]) * (InfoP->Mean[1]);
    InfoP->Cov[0][2] = (InfoP->Cov[0][2]/InfoP->Total_ObjSize) - 
	               (InfoP->Mean[0]) * (InfoP->Mean[2]);
    InfoP->Cov[1][0] = InfoP->Cov[0][1];
    InfoP->Cov[1][1] = (InfoP->Cov[1][1]/InfoP->Total_ObjSize) - 
	               (InfoP->Mean[1]) * (InfoP->Mean[1]);
    InfoP->Cov[1][2] = (InfoP->Cov[1][2]/InfoP->Total_ObjSize) - 
	               (InfoP->Mean[1]) * (InfoP->Mean[2]);
    InfoP->Cov[2][0] = InfoP->Cov[0][2];
    InfoP->Cov[2][1] = InfoP->Cov[1][2];
    InfoP->Cov[2][2] = (InfoP->Cov[2][2]/InfoP->Total_ObjSize) - 
	               (InfoP->Mean[2]) * (InfoP->Mean[2]);

}


void Jacobi_Rotate(SymMatrixP_In, Row_r, Col_p, Row_s, Col_q, sin_angle, tau)
/*--------------------------------------------------------------------------*/
/* purpose:                                                                 */ 
/*   it updates elements (Row_r, Col_p) and (Row_s, Col_q) for r!=p and s!=q*/
/* after applying a Jacobi plane rotation. Note that the same elements of   */
/* input array are changed                                                  */
/*--------------------------------------------------------------------------*/
/*Input and output variable definitions                                     */
/*--------------------------------------------------------------------------*/
double      **SymMatrixP_In;
double      sin_angle;              /* Sinine of a rotation angle           */
double      tau;                    /* Defined as sin_angle/(1 + cos_angle) */
int         Row_r, Col_p;           /* Array element (r,p)                  */
int         Row_s, Col_q;           /* Array element (s,q)                  */


{
/*--------------------------------------------------------------------------*/
/*Local variable definitions                                                */
/*--------------------------------------------------------------------------*/
    double      tmp_rp, tmp_sq;

    tmp_rp = SymMatrixP_In[Row_r][Col_p];
    tmp_sq = SymMatrixP_In[Row_s][Col_q];
    
    SymMatrixP_In[Row_r][Col_p] = tmp_rp - sin_angle*(tmp_sq + tau*tmp_rp);
    SymMatrixP_In[Row_s][Col_q] = tmp_sq + sin_angle*(tmp_rp - tau*tmp_sq);
}



int Eigen_SymMatrix(SymMatrix_In, Matrix_N, EigenValue_Out, EigenVector_Out)
/*--------------------------------------------------------------------------*/
/* purpose:                                                                 */ 
/*   it finds eigen_ values and vectors of symmetric matrix using Jacobi    */
/* plane rotation. Note that elements of SymMatrix above diagnal are        */
/* destroyed                                                                */
/*--------------------------------------------------------------------------*/
/*Input and output variable definitions                                     */
/*--------------------------------------------------------------------------*/
double      **SymMatrix_In;
double      **EigenVector_Out;      /* Cols store normalized eigen_vectors  */
double      *EigenValue_Out;        /* Corresponding unordered eigen_values */
int         Matrix_N;               /* Dimension of a symmetric matrix      */

{
/*--------------------------------------------------------------------------*/
/*Local variable definitions                                                */
/*--------------------------------------------------------------------------*/
    double      ctan_2angle, tan_angle, sin_angle, cos_angle;
    double      tau;                /* Defined as sin_angle/(1+cos_angle)   */
    double      Sum_OffDiagnals, DiagDiff_ji, tmp;
    double      IterThreshold;
    int         i, j, k;
    int         NoIteration;        /* No. of iterations to find Eigens     */

/*--------------------------------------------------------------------------*/
/*Initialize matrixs and vectors                                            */
/*--------------------------------------------------------------------------*/
    for ( i=0; i< Matrix_N; i++) {
	EigenVector_Out[i][i] = 1.0;
	for ( j=i+1; j<Matrix_N; j++) {
	    EigenVector_Out[i][j] = 0.0;
	    EigenVector_Out[j][i] = 0.0;
	}
    }
    for ( i=0; i<Matrix_N; i++)
	EigenValue_Out[i] = SymMatrix_In[i][i];

/*--------------------------------------------------------------------------*/
/*Cyclic Jacobi transform: iteratively set off-diagnal elements to zeros    */
/*--------------------------------------------------------------------------*/
    NoIteration = 0;
    Sum_OffDiagnals = 0.0;
    for ( i = 0; i < Matrix_N; i++)
	for ( j = i+1; j < Matrix_N; j++)
	    Sum_OffDiagnals += fabs(SymMatrix_In[i][j]);
    
    while ( (NoIteration < 30) && (Sum_OffDiagnals > TINY_VALUE) ) {
	NoIteration ++;
	
	if ( NoIteration > 4 ) 
	    IterThreshold = 0.0;
	else
	    IterThreshold = 0.2*Sum_OffDiagnals/(Matrix_N*Matrix_N);
	
	for ( i = 0; i < Matrix_N; i++) {
	    for ( j = i+1; j < Matrix_N; j++) {
		tmp = 100.0*SymMatrix_In[i][j];

/*--------------------------------------------------------------------------*/
/*After 4 sweeps, skip the rotation if the off_diagnal element is small     */
/*--------------------------------------------------------------------------*/
		if ( (NoIteration > 4) 
		     && 
		     (fabs(EigenValue_Out[i])+tmp == fabs(EigenValue_Out[i]))
		     &&
		     (fabs(EigenValue_Out[j])+tmp == fabs(EigenValue_Out[j]))
		   )
		    SymMatrix_In[i][j] = 0.0;
		else if ( fabs(SymMatrix_In[i][j]) > IterThreshold ) {
		    DiagDiff_ji = EigenValue_Out[j] - EigenValue_Out[i];
		    if ( fabs(DiagDiff_ji) + tmp == fabs(DiagDiff_ji) )
			tan_angle = SymMatrix_In[i][j]/DiagDiff_ji;
		    else {
			ctan_2angle = 0.5*DiagDiff_ji/SymMatrix_In[i][j];
			tan_angle = 1./(fabs(ctan_2angle) + 
					sqrt(1.0 + pow(ctan_2angle, 2.0)));
			if ( ctan_2angle < 0.0 ) 
			    tan_angle = -tan_angle;
		    }
		    cos_angle = 1.0/sqrt(1.0 + pow(tan_angle, 2.0));
		    sin_angle = tan_angle*cos_angle;
		    tau = sin_angle/ ( 1.0 + cos_angle );
		    
		    tmp = tan_angle*SymMatrix_In[i][j];
		    EigenValue_Out[i] -= tmp;
		    EigenValue_Out[j] += tmp;
		    SymMatrix_In[i][j] = 0.0;

/*--------------------------------------------------------------------------*/
/* Update upper off-diagnal elements after a rotation                       */
/*--------------------------------------------------------------------------*/
		    for ( k=0; k<i; k++) {
			Jacobi_Rotate(SymMatrix_In, k, i, k, j, 
				      sin_angle, tau);
		    }
		    for ( k=i+1; k<j; k++)
			Jacobi_Rotate(SymMatrix_In, i, k, k, j,
				      sin_angle, tau);
		    for ( k=j+1; k<Matrix_N; k++)
			Jacobi_Rotate(SymMatrix_In, i, k, j, k, 
				      sin_angle, tau);
/*--------------------------------------------------------------------------*/
/* Update off-diagnal elements of EigenVector_Out after a rotation          */
/*--------------------------------------------------------------------------*/
		    for ( k=0; k<Matrix_N; k++) {
			Jacobi_Rotate(EigenVector_Out, k, i, k, j, 
				      sin_angle, tau);
		    }
		}
	    }
	}
	for ( i = 0, Sum_OffDiagnals = 0.0; i < Matrix_N; i++)
	    for ( j = i+1; j < Matrix_N; j++)
		Sum_OffDiagnals += fabs(SymMatrix_In[i][j]);
    }
    if ( NoIteration == 30 )
	printf("(Eigen_SymMatrix): not diagnalized yet !!!\n");
    return NoIteration;
}


void X_2Matrix(Matrix1_In, Matrix2_In, Matrix_Out, N, M, L)
/*--------------------------------------------------------------------------*/
/* purpose:                                                                 */ 
/*   it computes a multiplication of a double NXM and double MXL matrix     */
/*--------------------------------------------------------------------------*/
/*Input and output variable definitions                                     */
/*--------------------------------------------------------------------------*/
double      **Matrix1_In, **Matrix2_In, **Matrix_Out;
int         N, M, L;               /* Dimension of matrixs                  */

{
/*--------------------------------------------------------------------------*/
/*Local variable definitions                                                */
/*--------------------------------------------------------------------------*/
    double *Ptr_O;
    int    i, j, k;

    for ( k=0, Ptr_O = Matrix_Out[0]; k<N; k++) {
	for ( j=0; j<L; j++, Ptr_O++) {
	    *Ptr_O = 0.0;
	    for ( i=0; i<M; i++) {
		*Ptr_O += Matrix1_In[k][i]* Matrix2_In[i][j];
	    }
	}
    }

}


void T_Matrix(Matrix_In, Matrix_Out, N, M)
/*--------------------------------------------------------------------------*/
/* purpose:                                                                 */ 
/*   it computes a transpose of a double NXM matrix                         */
/*--------------------------------------------------------------------------*/
/*Input and output variable definitions                                     */
/*--------------------------------------------------------------------------*/
double      **Matrix_In, **Matrix_Out;
int         N, M;                  /* Dimension of matrixs                  */

{
/*--------------------------------------------------------------------------*/
/*Local variable definitions                                                */
/*--------------------------------------------------------------------------*/
    int i, j;
    
    for ( i=0; i<N; i++) {
	for ( j=0; j<M; j++ ) {
	    Matrix_Out[j][i] = Matrix_In[i][j];
	}
    }

}
