/*-------------------------------------------------------------------
image_inc.c                                                       

Programmer       : Michelle Yan                                   
Revisions        : Paul Hughett                                   

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

   The source code contains some of the basic functions for        
image processing, including the histogram.                        
                                                                   
The included functions are                                      
   smoothing      : smooth a NXM image by one of selected filters  
   Scale2Char     : scale a float NXM array to a float array within
                    [0, 255], return maximum of original image.    
   Scale2Char3D   : scale a float NXMXL array to a 3D unsigned char
                    array, return maximum of original image.       
   Threshold2Bin  : threshold a float NXM image to char OBJECT &   
                    BACKGROUND using low & high threshold          
   histogram3D    : compute histogram for unsigned char NXMXL array
   select_threshold : automatically select an optimal threshold    
                      from histogram according to certain criterion
   refine_threshold : find a better threshold based on a given     
                      threshold. It handles the cases where two    
                      peaks in histogram are not comparable.       
   MeanStd        : compute a sample mean & standrad deviation from
                    a upper portion of histogram                   
   BinaryImg3D    : generate a 3D binary image using histogram3D,  
                    select_threshold & Threshold2Bin.              
   histo_mean     : return a mean ratio of two regions from a      
                    normalized histogram.                          
-------------------------------------------------------------------*/

static const char rcsid[] = "$Id: image_inc.c,v 1.22 2000/06/05 17:17:59 hughett Exp $ Release $Name: segm-0-90 $";

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <limits.h>

#include "common.h"
#include "config.h"
#include "definition.h"
#include "errm.h"
#include "memm.h"
#include "find_mask.h"
#include "image_inc.h"
#include "image_io.h"


/*-------------------------------------------------------------------
smoothing - Smooth a image with Gaussian or boxcar average  
      
Smooth a NXM image by a filter selected from the following list:

    No smoothing ........................... 0
    3X3 average filter  .................... 1                     
    3X3 Gaussian filter (PENNPET blur) ..... 2                     
    5X5 average filter  .................... 3                     
    5X5 Gaussian filter .................... 4                     
    7X7 Gaussian filter .................... 5                     

The borders of the image are unchanged, so that only the central
(N-2)*(M-2) or (N-4)*(M-4) or (N-6)*(M-6) array is smoothed. The
parameters of the Gaussian filter are selected such that standard
deviation is 0.85, 1.5 and 2.2 (pixel) for 3X3 or 5X5 or 7X7
respectively.

-------------------------------------------------------------------*/

void
smoothing( float ** InImgP,	/* Pointer to input & output images*/
	   float ** OutImgP,
	   int N,		/* Number of rows and columns      */ 
	   int M,
	   int FilterType )     /* Filter type, from list above    */
{

    float    *Ptr_in, *Ptr_out;
    int      i, j;

    /* Copy the input image to the output */
    /* FIXME:  Faster to just copy the borders */
    for (i=0, Ptr_in=InImgP[0], Ptr_out=OutImgP[0]; i<N*M; 
	 Ptr_in++, Ptr_out++, i++)
	*Ptr_out = *Ptr_in;

    /* Case 0:  No smoothing */
    if ( FilterType == NOSMOOTH ) {
       /* Do nothing */
    }

    /* Case 1:  3 x 3 average */
    else if ( FilterType == AVERAGE3X3 ) {
	for (i=1; i<N-1; i++)
	    for (j=1; j<M-1; j++) 
		OutImgP[i][j] = 
		    (InImgP[i-1][j-1] + InImgP[i-1][j] + InImgP[i-1][j+1] +
		     InImgP[i][j-1] + InImgP[i][j] + InImgP[i][j+1] +
		     InImgP[i+1][j-1] + InImgP[i+1][j] + InImgP[i+1][j+1])
			/ 9.;
    }

    /* Case 2:  3 x 3 Gaussian filter */
    else if ( FilterType == GAUSSIAN3X3 ) {
	for (i=1; i<N-1; i++)
	    for (j=1; j<M-1; j++) 
		OutImgP[i][j] = 
		    (0.25*(InImgP[i-1][j-1] + InImgP[i-1][j+1] +
			   InImgP[i+1][j-1] + InImgP[i+1][j+1]) +
		     0.5*(InImgP[i-1][j] +
			  InImgP[i][j-1] + InImgP[i][j+1] +
			  InImgP[i+1][j]) + 
		     InImgP[i][j]) / 4.0;
    }

    /* Case 3:  5 x 5 average filter */
    else if ( FilterType == AVERAGE5X5 ) {
	for (i=2; i<N-2; i++)
	    for (j=2; j<M-2; j++) 
		OutImgP[i][j] = 
		    (InImgP[i-2][j-2] + InImgP[i-2][j-1] + InImgP[i-2][j] +
		     InImgP[i-2][j+1] + InImgP[i-2][j+2] + 
		     InImgP[i-1][j-2] + InImgP[i-1][j-1] + InImgP[i-1][j] +
		     InImgP[i-1][j+1] + InImgP[i-1][j+2] + 
		     InImgP[i][j-2] + InImgP[i][j-1] + InImgP[i][j] +
		     InImgP[i][j+1] + InImgP[i][j+2] + 
		     InImgP[i+1][j-2] + InImgP[i+1][j-1] + InImgP[i+1][j] +
		     InImgP[i+1][j+1] + InImgP[i+1][j+2] + 
      		     InImgP[i+2][j-2] + InImgP[i+2][j-1] + InImgP[i+2][j] +
		     InImgP[i+2][j+1] + InImgP[i+1][j+2] ) / 25.;
    }

    /* Case 4:  5 x 5 Gaussian filter */
    else if ( FilterType == GAUSSIAN5X5 ) {
	for (i=2; i<N-2; i++)
	    for (j=2; j<M-2; j++) 
		OutImgP[i][j] = 
		    (0.16777216*(InImgP[i-2][j-2] + InImgP[i-2][j+2] + 
				 InImgP[i+2][j-2] + InImgP[i+2][j+2]) +
		    0.32768*(InImgP[i-2][j-1] + InImgP[i-2][j+1] + 
			     InImgP[i-1][j-2] + InImgP[i-1][j+2] + 
			     InImgP[i+1][j-2] + InImgP[i+1][j+2] + 
			     InImgP[i+2][j-1] + InImgP[i+2][j+1]) + 
	            0.4096*(InImgP[i-2][j] +
			    InImgP[i][j-2] + InImgP[i][j+2] + 
			    InImgP[i+2][j]) +
		    0.64*(InImgP[i-1][j-1] + InImgP[i-1][j+1] + 
			  InImgP[i+1][j-1] + InImgP[i+1][j+1]) + 
		    0.8*(InImgP[i-1][j] + 
                         InImgP[i][j-1] + InImgP[i][j+1] + 
			 InImgP[i+1][j]) +
		    InImgP[i][j]) / 11.69092864;
    }

    /* Case 5:  7 x 7 Gaussian filter */
    else if ( FilterType == GAUSSIAN7X7 ) {
	for (i=3; i<N-3; i++)
	    for (j=3; j<M-3; j++) 
		OutImgP[i][j] = 
		    (0.150094635*(InImgP[i-3][j-3] + InImgP[i-3][j+3] + 
				 InImgP[i+3][j-3] + InImgP[i+3][j+3]) +
		    0.254186583*(InImgP[i-3][j-2] + InImgP[i-3][j+2] + 
				 InImgP[i-2][j-3] + InImgP[i-2][j+3] +
				 InImgP[i+2][j-3] + InImgP[i+2][j+3] +
				 InImgP[i+3][j-2] + InImgP[i+3][j+2]) +
		    0.34867844*(InImgP[i-3][j-1] + InImgP[i-3][j+1] + 
				InImgP[i-1][j-3] + InImgP[i-1][j+3] + 
				InImgP[i+1][j-3] + InImgP[i+1][j+3] + 
				InImgP[i+3][j-1] + InImgP[i+3][j+1]) + 
		    0.387420489*(InImgP[i-3][j] + 
				 InImgP[i][j-3] + InImgP[i][j+3] + 
				 InImgP[i+3][j]) + 
		    0.43046721*(InImgP[i-2][j-2] + InImgP[i-2][j+2] + 
				InImgP[i+2][j-2] + InImgP[i+2][j+2]) +
		    0.59049*(InImgP[i-2][j-1] + InImgP[i-2][j+1] + 
			     InImgP[i-1][j-2] + InImgP[i-1][j+2] + 
			     InImgP[i+1][j-2] + InImgP[i+1][j+2] + 
			     InImgP[i+2][j-1] + InImgP[i+2][j+1]) + 
	            0.6561*(InImgP[i-2][j] +
			    InImgP[i][j-2] + InImgP[i][j+2] + 
			    InImgP[i+2][j]) +
		    0.81*(InImgP[i-1][j-1] + InImgP[i-1][j+1] + 
			  InImgP[i+1][j-1] + InImgP[i+1][j+1]) + 
		    0.9*(InImgP[i-1][j] + 
                         InImgP[i][j-1] + InImgP[i][j+1] + 
			 InImgP[i+1][j]) +
		    InImgP[i][j]) / 23.88316952;
    }

    return;
}


/*-------------------------------------------------------------------
Scale2Char - Rescale image to range [0,255]

This function determines the min and max gray values in the input
array.  If both are between 0 and 255, then the gray values are
rounded to the nearest integer.  Otherwise, the gray values are
multiplied by 255 divided by the max value and rounded to the nearest
integer, except that negative values are set to zero.  The original
maximum value found is returned.  (FIXME?) The minimum and maximum
intensity values found are printed on standard output.
-------------------------------------------------------------------*/

float
Scale2Char( float ** InImgP,	/* (In) Image to be scaled         */
	    float ** OutImgP,   /* (Out) Scaled image              */
	    int N,		/* Number of rows and columns      */
	    int M )
{

    float *Ptr_1, *Ptr_2;
    float Ratio, Max_val, Min_val;
    int   i;

    /* Determine the min and max values in the image */
    Max_val = InImgP[0][0];
    Min_val = InImgP[0][0];
    for ( i=0, Ptr_1=InImgP[0]; i< N*M; Ptr_1++, i++) {
	if ( (*Ptr_1) < Min_val )
	    Min_val = *Ptr_1;
	else if ( (*Ptr_1) > Max_val ) 
	    Max_val = *Ptr_1;
    }
    printf("\nMaximum and Minimum values are: %f %f \n",
	   Max_val, Min_val);

    /* Now rescale to the range [0,255] and round to integers */
    Ptr_1 = InImgP[0];
    Ptr_2 = OutImgP[0];
    if ( (Min_val >= 0.0) && (Max_val <= 255.0) ) {
	for (i=0; i<N*M; Ptr_1++, Ptr_2++, i++)
	    *Ptr_2 = (int)((*Ptr_1) + 0.5);
    }
    else {
	Ratio = 255.0/Max_val;
	for ( i=0; i< N*M; Ptr_1++, Ptr_2++, i++) 
	    if ( (*Ptr_1) < 0.0 ) 
		*Ptr_2 = 0.0;
	    else
		*Ptr_2 = (int)((*Ptr_1) * Ratio + 0.5);
    }
    return Max_val;
}


/*-------------------------------------------------------------------
Scale2Char3D - Smooth, scale, and quantize a volume image

Optionally smooth the input image then scale and quantize to within the
range [0,255].  If the image is already within the range [0,255], do
not rescale it.  The smoothing options are those defined for the
function smoothing.  The input image is replaced by the smoothed
input image.  The upper limit 255 is disguised under the definitint
HISTO_L - 1 for historical reasons.  (FIXME?) The maximum and minimum
intensity values found are printed on standard output.
---------------------------------------------------------------------*/

float
Scale2Char3D( float *** InImg,             /* Input image */
	      unsigned char *** OutChar,   /* Output image */
	      int N,                       /* Image dimensions */
	      int M,
	      int L,
	      int SmoothOpt )               /* Smoothing option */
{
    float         **tmpImg, *ptr_f;
    float         Ratio, Max_val, Min_val;
    int           i, k;
    unsigned char *ptr_c;

    /* Optionally smooth the input image */
    if ( SmoothOpt != NOSMOOTH ) {
	tmpImg = alloc_img(N, M);
	for ( i=0; i<L; i++ ) {
	    smoothing(InImg[i], tmpImg, N, M, SmoothOpt);
	    copy_matrix (tmpImg, InImg[i], N, M);
	}
	free_img (tmpImg);
    }

    /* Determine minimum & maximum intensity value */
    Max_val = InImg[0][0][0];
    Min_val = InImg[0][0][0];
    for ( k=0; k<L; k++ ) {
	for ( i=0, ptr_f=InImg[k][0]; i<N*M; i++, ptr_f++ ) {
	    if ( *ptr_f < Min_val )
		Min_val = *ptr_f;
	    else if ( *ptr_f > Max_val ) 
		Max_val = *ptr_f;
	}
    }
    printf("\nMaximum & Minimum values = (%f, %f)\n",
	   Max_val, Min_val);

    /* Scale a 3D image to [0, HISTO_L-1] by the maximum value */
    if ( Max_val < (float) HISTO_L ) {
	for ( k=0; k<L; k++ ) {
	    ptr_c = OutChar[k][0];
	    ptr_f = InImg[k][0];
	    for ( i=0; i<N*M; i++, ptr_c++, ptr_f++ ) {
		if ( *ptr_f < 0.0 ) 
		    *ptr_c = 0;
		else 
		    *ptr_c = (int) (*ptr_f + 0.5);
	    }
	}
    }
    else if ( Max_val >= (float) HISTO_L ) {
	Ratio = (float) (HISTO_L-1)/Max_val;
	for ( k=0; k<L; k++ ) {
	    ptr_c = OutChar[k][0];
	    ptr_f = InImg[k][0];
	    for ( i=0; i<N*M; i++, ptr_c++, ptr_f++ ) {
		if ( *ptr_f < 0.0 ) 
		    *ptr_c = 0;
		else 
		    *ptr_c = (int) (*ptr_f * Ratio + 0.5);
	    }
	}
    }

    return Max_val;
}
	    

/*--------------------------------------------------------------------
Threshold2Bin - Binarize an image by thresholding 

This function binarizes a 2D image by thresholding.  If an input pixel
is between TLow and THigh inclusive, the output pixel is set to
OBJECT, otherwise to BACKGROUND.  The number of OBJECT pixels found
is returned.
--------------------------------------------------------------------*/

int
Threshold2Bin(
   float **InImgP,            /* Input gray scale image */
   unsigned char **OutChar,   /* Output binary image */
   float TLow,	              /* Lower and upper thresholds */
   float THigh,
   int N,		      /* Number of rows */ 
   int M )		      /*  and columns   */
{
    int           Obj_Size = 0;   /*Numbers of pixels > threshold    */
    int           i;
    float         *ptr_f;
    unsigned char *ptr_c;

    /* Loop over the image, processing each pixel */
    ptr_f = InImgP[0];
    ptr_c = OutChar[0];
    for (i=0; i<N*M; ptr_f++, ptr_c++, i++) {
        if ( *ptr_f >= TLow && *ptr_f <= THigh ) {
	    *ptr_c = OBJECT;
	    Obj_Size ++;
	}
	else
	    *ptr_c = BACKGROUND;
    }

    return Obj_Size;
}


/*-------------------------------------------------------------------
histogram3D - Compute normalized histogram of 3D unsigned char image

This function computes the histogram of a 3D unsigned char image,
normalized so that the entries corresponding to non-zero gray values
sum to one.  The total number of non-zero pixels is returned.  The
histogram array provided by the user must be 256 elements long, one
more than the maximum possible value of an unsigned char.  It is a
fatal error if the image is entirely zero.

FIXME:  This code could break nastily on a machine for which unsigned
char is more than 8 bits, a thing not prohibited by the standard.
--------------------------------------------------------------------*/

int
histogram3D(
   unsigned char *** InChar,   /* Pointer to input image    */
   float * OutHistoP,	       /* Pointer to 1D histogram   */
   int N,                      /* Dimensions of 3D image    */
   int M,
   int L )
{
    float         *ptr_f;
    int           i, k, Total_PixNo = 0;
    unsigned char *ptr_c;

    /* Initialize a histogram vector of length HISTO_L = 256 */
    ptr_f = OutHistoP;
    for ( i = 0; i < HISTO_L; ptr_f++, i++)
	*ptr_f = 0.;

    /* Compute an unnormalized histogram */
    for ( k=0; k<L; k++ ) {
	for ( i=0, ptr_c=InChar[k][0]; i<N*M; ptr_c++, i++) {
	    OutHistoP[(*ptr_c)] ++;
	    if ( *ptr_c != 0 )
		Total_PixNo ++;
	}
    }
    if ( Total_PixNo == 0 ) 
	fatal("hisogram3D: input image is ZERO.");

    /* Normalize histogram without zero pixels */
    ptr_f = OutHistoP + 1;
    for ( i=1; i<HISTO_L; ptr_f++, i++) {
	if ( *ptr_f != 0.0 ) {
	    (*ptr_f) /= (float)Total_PixNo;
	}
    }

    return Total_PixNo;
}


/*--------------------------------------------------------------------
Kmeans_threshold - Use K-means algorithm to compute class thresholds

Given the gray-level histogram of an image, this function uses the
K-means algorithm to compute thresholds for dividing the image into a
given number of classes.  The array of thresholds is one element longer
than the number of classes, since it must list both the lower and upper 
threshold for each class.  The lower limit is inclusive; the upper limit
is exclusive.  That is, class n consists of those pixels with gray level
in the half-open interval [thresh[i], thresh[i+1]).  The caller is
responsible for allocating the threshold array.  Some of the classes
will be empty if there are fewer distinct gray values than classes;
in this case, the position of the thresholds is not well defined.

This function serves the same general purpose as select_threshold,
but uses a different algorithm and finds different values.  The
two should not be assumed to be equivalent without investigation.

For simplicity, memory is allocated on the stack for a given maximum
number of classes rather than gotten from malloc.  The maximum can
be made as large as desired, although malloc should probably be used
instead if it exceeds 100 or so.
-------------------------------------------------------------------*/

#define MAXCLASS 20          /* Maximum number of classes */
#define MAXCHG   0.2         /* Stop when means change by less than this */

void
Kmeans_threshold(
    float *histogram,         /* Gray-level histogram */
    int ngray,                /* Number of gray levels in histogram */
    int nclass,               /* Number of classes to create */
    int *thresh)              /* (out) array of thresholds */
{   
    float means[MAXCLASS];    /* Class means */
    double maxchg;            /* Largest change in a class mean */
    int iclass, igray;
    double sumpg, sump, new;

    /* Verify that arguments are reasonable */
    if (ngray <= 0) {
        fprintf(stderr, "Bad arg ngray in Kmeans_threshold\n");
	exit(EXIT_FAILURE); }
    if (nclass <= 0 || nclass >= ngray) {
        fprintf(stderr, "Bad arg nclass in Kmeans_threshold\n");
	exit(EXIT_FAILURE); }

    /* Create an initial set of class means and thresholds */
    for (iclass = 0; iclass < nclass; iclass++)
        means[iclass] = ((iclass + 0.5)*ngray)/nclass;
    thresh[0]      = 0;
    thresh[nclass] = ngray;
    for (iclass = 1; iclass < nclass; iclass++)
        thresh[iclass] = (means[iclass-1] + means[iclass])/2.0;

    /* Repeatedly recompute thresholds until class means are stable */
    for (maxchg = 0.0; maxchg > MAXCHG; maxchg = 0.0) {

        /* Classify gray levels; recompute means and thresholds */
        for (iclass = 0; iclass < nclass; iclass++) {
	    sumpg = sump = 0.0;
	    for (igray = thresh[iclass]; igray < thresh[iclass+1]; igray++) {
	        sumpg += igray*histogram[igray];
		sump  += histogram[igray];
	    }
	    if (sump > 0)
		new = sumpg / sump;
	    else      /* Avoid zero-divide if class is empty */
		new = (thresh[iclass] + thresh[iclass+1])/2.0;
	    if (fabs(new - means[iclass]) > maxchg)
	        maxchg = fabs(new - means[iclass]);
	    means[iclass] = new;
	    if (iclass > 0)
               thresh[iclass] = (means[iclass-1] + means[iclass])/2.0;
	}
    }

    return;
}



/*--------------------------------------------------------------------
select_threshold - Compute optimal threshold from a histogram

This function examines a histogram (such as generated by histogram3D)
and chooses the optimal threshold to maximize the separability of the
two classes, or to separate an object from the background.  (It's not
clear which it is intended to do.)  It is also not clear whether you
test on < threshold or <= threshold to separate the classes.  If you
want to separate more than two classes, this function won't help you.
The optimal threshold is returned.  The histogram array is assumed to
be HISTO_L = 256 elements long.

The optimality criterion is as follows.  Let M be the mean intensity
of all the non-zero pixels and m(i) the mean intensity of background
pixels as a function of the threshold i; let p(i) be the fraction of
pixels assigned to the background.  Then the optimal threshold is the
value of i that corresponds to the first maximum of Sigma(i) =
(M-m(i))^2 * p(i) / (1-p(i)).  The rationale behind this criterion is
obscure, to say the least.
----------------------------------------------------------------------*/

int
select_threshold(               /* Return the optimal threshold */
   float * Ptr_Histo1D )	/* Pointer to input of 1D histogram*/
{
    float   tmpval;
    float   *ptr_tmp1;
    float   Mean_Total = 0.0;   /* Average intensity of non-zero pixels */
    float   Mean_1 = 0.0;       /* Weight sum of background pixel intensity */
    float   Prob_1 = 0.0;       /* Fraction of pixels in background */
    float   Sigma2_old, Sigma2_new;  /* Optimality cost function */
    int     i, opt_threshold;
    int     First_nozero;       /* First intensity with pixel count > 0 */

    /* Optionally print histogram for debugging */
    if (0) {
        fprintf(stderr, "    ---- ");
        for (i = 1; i < HISTO_L; i++) {
	    fprintf(stderr, "%8.2f ", 100*Ptr_Histo1D[i]);
	    if (i%8 == 7) fprintf(stderr, "\n");
	}
    }

    /* Regard a normalized histogram as a probability distribution and
       compute the total mean */
    ptr_tmp1 = Ptr_Histo1D + 1;
    First_nozero = 1;
    while ( (First_nozero < HISTO_L) && ( (*ptr_tmp1) == 0.) ) {
	First_nozero ++;
	ptr_tmp1 ++;
    }
    if ( First_nozero == HISTO_L ) {
	fatal("<<< ERROR (select_threshold) <<<: zero histogram.");
    }

    ptr_tmp1 = Ptr_Histo1D + First_nozero;
    for ( i = First_nozero; i < HISTO_L; ptr_tmp1++, i++)
	if ( *ptr_tmp1 != 0. ) {
	    Mean_Total += (float)(i) * (*ptr_tmp1);
	}

    /* Search an optimal threshold from histogram:                    */
    /* Compute Probability and Mean of region 1 at new opt_threshold  */
    /* and compare Sigma2_new to Sigma2_old                           */
    opt_threshold = First_nozero - 1;
    Sigma2_old = -1.0;
    Sigma2_new = 0.;
    while ( (opt_threshold < HISTO_L) && (Sigma2_new >= Sigma2_old) ) {
	opt_threshold ++;

	Prob_1 += Ptr_Histo1D[opt_threshold];
	Mean_1 += (float)(opt_threshold) * Ptr_Histo1D[opt_threshold];
	Sigma2_old = Sigma2_new;
	tmpval = Mean_Total * Prob_1 - Mean_1;
	Sigma2_new = tmpval * tmpval / (Prob_1 * (1. - Prob_1));
	if (1)
	    fprintf(stderr, "Opt_threshold & Sigma2_new =: %d %f \n",             
	        opt_threshold, Sigma2_new);
    }
    opt_threshold --;

    return opt_threshold;
}


/*-------------------------------------------------------------------
refine_threshold - Find a better threshold for certain images

This function selects a possibly better threshold based on CurrectT.      
It is for handling the cases that two peaks in histogram are not  
comparable.  The algorithm used is not clear, but appears to have
something to do with dilating the original thresholded image, using
it as a mask for the original image, and taking the histogram of the
masked image.
-------------------------------------------------------------------*/

int
refine_threshold( float Threshold,
		  unsigned char ***MaskIn,   /* Mask image */
		  unsigned char ***CharIn,   /* Quantized image */
		  int N,	             /* Number of rows, */
		  int M,	             /* columns, */
		  int L )	             /* and slices */

{
    float         *Sml_Histo;
    int           i, k, opt_threshold;
    unsigned char ***tmpMask_1, ***tmpMask_2, 
                  *ptr_c1, *ptr_c2, *ptr_c3,
                  ***ptr3_1, ***ptr3_2;

    /* Allocate memory  */
    Sml_Histo = (float *) callock( HISTO_L, sizeof(float));
    tmpMask_1 = alloc_3Dchar(N, M, L);
    tmpMask_2 = alloc_3Dchar(N, M, L);

    /* Based on input Threshold, get a binary image and dilate it */
    for ( k=0; k<L; k++ ) { 
	ptr_c1 = tmpMask_1[k][0];
	ptr_c2 = CharIn[k][0];
	ptr_c3 = MaskIn[k][0];
	for ( i=0; i<N*M; ptr_c1++, ptr_c2++, ptr_c3++, i++) {
	    if ( *ptr_c3 != BACKGROUND ) {
		if ( *ptr_c2 > Threshold )
		    *ptr_c1 = BACKGROUND;
		else
		    *ptr_c1= OBJECT;
	    }
	    else
		*ptr_c1 = BACKGROUND;
	}
    }
    dilate3D( tmpMask_1, tmpMask_2, N, M, L );
    for ( k=0, ptr3_1=tmpMask_1, ptr3_2=tmpMask_2; k<L; 
	  k++, ptr3_1++, ptr3_2++ ) {
	dilate3D (ptr3_2, ptr3_1, N, M, 1);
    }
    free_3Dchar (tmpMask_2, L);

    for ( k=0; k<L; k++ ) { 
	ptr_c1 = tmpMask_1[k][0];
	ptr_c2 = CharIn[k][0];
	ptr_c3 = MaskIn[k][0];
	for ( i=0; i<N*M; ptr_c1++, ptr_c2++, ptr_c3++, i++) {
	    if ( *ptr_c1 == OBJECT && *ptr_c3 != BACKGROUND ) 
		*ptr_c1 = *ptr_c2;
	    else
		*ptr_c1 = 0;
	}
    }

    /* Select a threshold from 3D histogram of dilated OBJECT            */
    histogram3D(tmpMask_1, Sml_Histo, N, M, L);
    opt_threshold = select_threshold(Sml_Histo);

    free ( Sml_Histo );
    free_3Dchar (tmpMask_1, L);

    return opt_threshold;
}


/*-------------------------------------------------------------------
MeanStd - Compute mean and standard deviation of upper class

Given a gray-level histogram and a threshold, this function computes
the mean and standard deviation of the portion of the histogram lying
above the threshold.  More precisely, the portion lying at or above the
integer part of the threshold.  The given threshold will normally be
an optimal threshold computed by select_threshold, but this is not
required.
--------------------------------------------------------------------*/

void
MeanStd( float *Histo,		/* Pointer to input of 1D histogram */
	 float OptT,		/* (Optimal) Threshold */
	 float *Mean_O,		/* Mean for output */
	 float *STD_O )		/* STD for output */

{
  float tmpval;
  float     *ptr_f, mean = 0.0, std = 0.0, prob = 0.0;
  int       i;


  ptr_f = Histo + (int) OptT;
  for ( i=(int)OptT; i<HISTO_L; i++, ptr_f++ ) {
    prob += *ptr_f;
    mean += i * (*ptr_f);
  }
  if ( prob > ZERO ) { 
    mean /= prob;

    ptr_f = Histo + (int) OptT;
    for ( i=(int)OptT; i<HISTO_L; i++, ptr_f++ ) {
      tmpval = i - mean;
      std += tmpval * tmpval * (*ptr_f);
    }
    std = sqrt (std/prob);
  }
  else {
    mean = 0.0;
    std = 0.0;
  }

  *Mean_O = mean;
  *STD_O = std;
}



/*--------------------------------------------------------------------------
BinaryImg3D - Binarize a 3D gray-level image by automatic thresholding

This function converts a 3D gray-level image into a binary mask image
with values OBJECT and BACKGROUND by thresholding at level(s)
determined by examination of the gray level threshold.  The lower
threshold is determined by select_threshold; if UpperT_YN is equal to
"y", the upper threshold is set to 2 standard deviations above the
mean of the gray levels lying above the lower threshold.  Pixels between
the lower and upper thresholds are set to OBJECT; the rest are set to
BACKGROUND.
---------------------------------------------------------------------------*/

void
BinaryImg3D(
   unsigned char ***CharIn,     /* Input gray level image */
   unsigned char ***MaskOut,    /* Output mask image */
   int N,		        /* Rows, columns and slices */
   int M,
   int L,
   char *UpperT_YN )	         /* Indicator for upper threshold */

{
  float              **tmpImg, *HistoP;
  float              OptT, spl_mean, spl_std, UpperT, *ptr_f;
  int				k;
  int			i;
  unsigned char	*ptr_c;


  HistoP = (float *) callock(HISTO_L, sizeof(float));
  tmpImg = alloc_img( N, M );
    
  /*  Compute histogram from a 3D image */
  histogram3D( CharIn, HistoP, N, M, L );

  /*  Select an optimal Threshold from histogram to generate a binary image */
  OptT = (float) select_threshold( HistoP );
  printf("\nOptimal Threshold for 3D image = %f\n", OptT);
  if ( strcmp( UpperT_YN, "y" ) != 0 ) { 
    UpperT = 255.0;
  }
  else {
    MeanStd (HistoP, OptT, &spl_mean, &spl_std);
    UpperT = spl_mean + 2.0 * spl_std;
    printf("Mean, STD & UpperT =: %f, %f, %f\n", spl_mean, spl_std, UpperT);
    UpperT = min(255.0, UpperT);
  }

  for ( k=0; k<L; k++ ) { 
    ptr_f = tmpImg[0];
    ptr_c = CharIn[k][0];
    for (i = N * M; i > 0; i--) {
      *ptr_f++ = *ptr_c++;
    }
    Threshold2Bin( tmpImg, MaskOut[k], OptT, UpperT, N, M );
  }

  free (HistoP);
  free_img (tmpImg);
}


/*-----------------------------------------------------------------------
Kmeans_binarize - Binarize a uchar image based on Kmeans segmentation

This function converts a 3D unsigned char gray scale image into a
binary image using thresholds derived from a K-means segmentation of
the histogram into some specified number of classes.  Pixels with gray
levels in a specified number of lower and upper classes are translated
to BACKGROUND; all other pixels are translated to OBJECT.  Pixels with
gray value zero are always lumped into the first class, which may be
classified as either BACKGROUND or OBJECT, depending on the value of
nlower.  (FIXME:  Is this the right behavior?)

This function serves much the same purpose as does BinaryImg3D but uses
a different algorithm and produces different results.  The two should
not be considered equivalent without investigation.
-------------------------------------------------------------------------*/

void
Kmeans_binarize(
    unsigned char ***CharIn,     /* Input gray level image */
    unsigned char ***MaskOut,    /* Output mask image */
    int N,		         /* Rows, columns and slices */
    int M,
    int L,
    int nclass,                  /* Nr of classes for segmentation */
    int nlower,                  /* Nr of lower classes as background */
    int nupper)                  /* Nr of upper classes as background */
{
    float *histogram;            /* Image histogram */
    int *thresh;                 /* Class thresholds */
    long count, i;               /* Nr of pixels per slice */
    int lower, upper;            /* Lower and upper gray values for OBJECT */
    int maxgray = 256;           /* Max possible unsigned char gray value */
    int k;
    unsigned char *pc, *pm;

    /* Verify that arguments are reasonable */
    if (nclass <= 0 || nclass >= 256) 
	fatal("Bad arg nclass in Kmeans_binarize");
    if (nlower < 0 || nlower >= nclass)
	fatal("Bad arg nlower in Kmeans_binarize");
    if (nupper < 0 || nupper >= nclass)
	fatal("Bad arg nupper in Kmeans_binarize");

    /* Compute histogram for the image; ignore zero gray values */
    histogram = mallock(maxgray*sizeof(float));
    histogram3D(CharIn, histogram, N, M, L);
    histogram[0] = 0.0;

    /* Select thresholds via K-means algorithm */
    thresh = mallock((nclass+1)*sizeof(int));
    Kmeans_threshold(histogram, maxgray, nclass, thresh);
    if (1) {
	fprintf(stderr, "K-means thresholds are: ");
	for (k = 0; k <= nclass; k++)
	    fprintf(stderr, "%d ", thresh[k]);
	fprintf(stderr, "\n");
    }

    /* Binarize the image using the selected thresholds */
    /* FIXME:  Should this be a separate routine? */
    count = (long)N*M;
    lower = thresh[nlower];
    upper = thresh[nclass-nupper];
    fprintf(stderr, "Object thresholds are %d, %d\n", lower, upper);
    for (k = 0; k < L; k++) {
	pc = &CharIn [k][0][0];
	pm = &MaskOut[k][0][0];
	for (i = 0; i < count; i++) {
	    if (lower <= pc[i] && pc[i] < upper) pm[i] = OBJECT;
	    else                                 pm[i] = BACKGROUND;
	}
    }

    /* Free memory and return */
    free(thresh);
    free(histogram);
    return;
}


/*-------------------------------------------------------------------
histo_mean - Compute mean ratio from a histogram

Compute the mean ratio according to opt_threshold of a normalized 
histogram.  No one seems to know what this means or what it is
useful for.  Fortunately, this function does not appear to be used
anywhere, wither.                                                   
-------------------------------------------------------------------*/

float
histo_mean( float *HistoP,
	    float opt_threshold )
{

    float   *Ptr_1;
    float   Mean_Total = 0.0, Mean_1 = 0.0, Prob_1 = 0.0;
    float   Mean_2, CmptRatio_21;
    int     i, First_nozero;


   /*Regard a normalized histogram as a probability distribution and    */
   /*compute the total mean                                             */
    Ptr_1 = HistoP + 1;
    First_nozero = 1;
    while ( (First_nozero < HISTO_L) && ( (*Ptr_1) == 0.0) ) {
	First_nozero ++;
	Ptr_1 ++;
    }
    if ( First_nozero == HISTO_L ) {
	fatal("<<< ERROR (select_threshold) <<<: zero histogram.");
    }

    Ptr_1 = HistoP + First_nozero;
    for ( i = First_nozero; i < HISTO_L; Ptr_1++, i++)
	if ( *Ptr_1 != 0.0 ) {
	    Mean_Total += (float)(i) * (*Ptr_1);
	}

   /*Compute Probability and Mean of region 1 and 2 according to        */
   /*opt_threshold                                                      */
    for ( i = First_nozero; i <= (int) opt_threshold; i++) {
	Prob_1 += HistoP[i];
	Mean_1 += (float)(i) * HistoP[i];
    }
    Mean_1 /= Prob_1;
    Mean_2 = (Mean_Total - Prob_1*Mean_1) / (1-Prob_1); 

    if ( Mean_1 > TINY_VALUE ) 
	CmptRatio_21 = Mean_2/Mean_1;
    else
	CmptRatio_21 = INFINITE;

    return CmptRatio_21;
}
