/*----------------------------------------------------------------------
list.c - Functions for manipulating stacks and linked lists. 

Author:     Michelle Yan
Revisions:  Dan Widyono, Paul Hughett

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.
----------------------------------------------------------------------*/

static const char rcsid[] = "$Id: list.c,v 1.9 2000/06/05 17:17:59 hughett Exp $ Release $Name: segm-0-90 $";

#include "memm.h"
#include "list.h"



/*----------------------------------------------------------------------
                  Functions to manipulate stacks

The stack consists of a header of type stackType, which points to the
top and bottom of a singly linked list of nodes of type singleLinkType,
each of which points to a single item kept in the stack.  The header
also contains a pointer to the current node, which is used to scan
the stack without actually popping entries, and a count of the number
of items currently on the stack.

New headers and nodes are created automatically by calling callock,
and returned automatically by calling free.  The items pointed to
by the stack must be allocated and returned by the caller.

FIXME:  Keep a free list of headers and nodes for more efficient
allocation.  malloc and free are not efficient for frequent small
allocations.

----------------------------------------------------------------------*/


/* LIST_newStack - Create a new, empty stack */
void *
LIST_newStack( void )
{
  void * this;

  this = callock( (size_t) 1, sizeof( stack_t ) );

  return this;
}


/* LIST_freeStack - Free the header and all nodes of a stack */
void
LIST_freeStack( stack_t * this )
{
  if (this)
  {
    void * point;

    while ((point = LIST_popStack( this )))
      free (point);
    free (this);
  }
}


/* LIST_pushStack - Push a new element onto the top of a stack */
void
LIST_pushStack( stack_t * thisStack,
		void * content )
{
  sLink * newLink;

  newLink = (sLink *) mallock(sizeof(sLink));
  newLink->next = thisStack->top;
  newLink->content = content;
  thisStack->top = newLink;
  thisStack->stackSize ++;
}


/* LIST_popStack - Pop the top element off a stack */
void *
LIST_popStack( stack_t * thisStack )
{
  if (thisStack->stackSize > 0)
  {
    sLink * top;
    void * content;

    top = thisStack->top;
    content = top->content;
    thisStack->top = top->next;
    thisStack->stackSize --;
    free (top);
    return content;
  }

  return (void *) 0;
}


/* LIST_resetStack - Move the current pointer to the top of the stack */
void
LIST_resetStack( stack_t * thisStack )
{
  thisStack->current = thisStack->top;
}


/* LIST_getstack - Return current element of a stack and advance to next */
void *
LIST_getStack( stack_t * thisStack )
{
  if (thisStack->current)
  {
    void * content;

    content = thisStack->current->content;
    thisStack->current = thisStack->current->next;
    return content;	
  }

  return (void *) 0;	
}




/*----------------------------------------------------------------------
                      Doubly linked lists

This data structure consists of a header which points to the top and
bottom of a doubly linked list of nodes, each of which contains a 
pointer to the actual element in the list.  The header also contains
a pointer to a current element, which is used to scan the list without
modifying it, a count of the number of elements in the list, and pointers
to two functions to compare list elements and to free a list element.
The elements are sorted from first to last in ascending order, as defined
by the comparison function.

The comparison takes pointers to two contents A and B and should return
-1 if A<B, 0 if A=B, and +1 if A>B.  See the functions floatcmp and
intcmp later in this file for examples.

Headers and nodes are allocated and freed automatically as needed.  The
content or payload must be allocated by the caller, but will be freed
automatically using the free list element function.

The node freeing function takes a pointer to the contents of a node
and is expected to free any memory allocated to the contents, or to
do nothing if the contents should not be freed.  See the functions
defaultFreeContent and nullFreeContent later in this file for
examples of node freeing functions.

FIXME:  mallock and free are called directly for each header or node
needed.  This is rather inefficient and it would be better to keep
a free list of nodes and headers.

----------------------------------------------------------------------*/


/* LIST_newList - Create a new doubly linked list */
void *                             /* Returns the list created */
LIST_newList(
   listelemcompfunc compare,       /* Function to compare two nodes */
   listelemfreefunc freeContent )  /* Function to free a node */
{
  list_t *this;

  this = (list_t *) mallock(sizeof(list_t));
  this->top = this->bottom = this->current = 0;
  this->listSize = 0;

  this->compare = compare;
  this->freeContent = freeContent;

  return this;
}


/* LIST_sizeList - Return number of elements on a doubly linked list */
int
LIST_sizeList( list_t * this )
{
  return this->listSize;
}


/* LIST_topList - Set current element to be the top (first) element */
void
LIST_topList( list_t * this )
{
  this->current = this->top;
}


/* LIST_bottomList - Set current element to be the bottom (last) element */
void
LIST_bottomList( list_t * this )
{
  this->current = this->bottom;
}


/* LIST_isTopList - Is the current element the top element? */
int
LIST_isTopList( list_t * this )
{
  return (this->current == this->top);
}


/* LIST_isBottomList - Is the current element the bottom element? */
int
LIST_isBottomList( list_t * this )
{
  return (this->current == this->bottom);
}


/* LIST_nextList - advance to next element, if any */
void
LIST_nextList( list_t * this )
{ 
  if (this->current != this->bottom)
    this->current = this->current->next;
}


/* LIST_priorList - backup to the previous element, if any */
void
LIST_priorList( list_t * this )
{ 
  if (this->current != this->top)
    this->current = this->current->prior;
}


/* LIST_contentList - return a pointer to the content of the current element */
void *
LIST_contentList( list_t * this )
{ 
  return this->current->content;
}


/* LIST_contentAndNextList - return current content and advance to next */
void *
LIST_contentAndNextList( list_t * this )
{ 
  void * content = this->current->content;

  if (this->current != this->bottom)
    this->current = this->current->next;
  return content;
}


/******
LIST_findList - Find list entry with a given key

This function searches a (sorted) doubly linked list for a node
matching a given key, as defined by the comparison function, and makes
that node the current node.  If there is no matching node, either the
preceding or following node will be made current, depending on the
direction of the search.  The value returned is -1, 0, or +1, and is
the result of comparing the node contents to the desired key.  The
search starts from the current node.  The results are undefined if the
list is not sorted.  The value -1 is returned if the list is empty.
******/


int
LIST_findList( list_t * this, void * ref )
{
  dLink *current;
  int order = -1;

  if (!(this->listSize))       /* empty list    */
    return -1;

  if ((order = (*this->compare)(this->top->content, ref)) >= 0)
  {
    this->current = this->top;
    return order;
  }

  if ((order = (*this->compare)(this->bottom->content, ref)) <= 0) 
  {
    this->current = this->bottom;
    return order;
  }

  current = this->current;

  if ((*this->compare)(current->content, ref) > 0)
  {
    do {
      current = current->prior;
    } while ((order=(*this->compare)(current->content,ref)) > 0);
    this->current = current;
    return order;
  }

  if ((*this->compare)(current->content, ref) < 0)
  {
    do {
      current = current->next;
    } while ((order=(*this->compare)(current->content,ref)) < 0);
    this->current = current;
    return order;
  }

  return 0;
}


/******
LIST_insertList - Add node to a sorted doubly linked list

This function adds a node in the right position of a sorted doubly 
linked list, as defined by the comparison function of that list.
If a node with the same key already exists, the new node might
be added either before or after the old one.
******/

void
LIST_insertList( list_t * this,
		 void * content )
{
  dLink *newLink=0;
  int order;

  newLink = (dLink *) mallock(sizeof(dLink));
  newLink->content = content;
  newLink->next = newLink->prior = 0;

  if (!(this->listSize))
  { this->top = this->bottom = this->current = newLink;
    this->listSize ++;
    return;
  }

  order = LIST_findList (this, content);
  if (order >= 0 && this->current == this->top)
  { newLink->next = this->top;
    this->top->prior = newLink;
    this->top = newLink;
  }
  else if (order <= 0 && this->current == this->bottom)
  { newLink->prior = this->bottom;
    this->bottom->next = newLink;
    this->bottom = newLink;
  }
  else if (order >= 0)
  { newLink->prior = this->current->prior;
    newLink->next = this->current;
    this->current->prior->next = newLink;
    this->current->prior = newLink;
  }
  else
  { newLink->next = this->current->next;
    newLink->prior = this->current;
    this->current->next->prior = newLink;
    this->current->next = newLink;
  }

  this->listSize ++;
  return;
}


/******
LIST_retrieveList - Remove current element from a doubly linked list

This function removes the current element from a doubly linked list
and returns the contents.  The memory allocated to the node is 
automatically reclaimed; the memory allocated to the contents is
unchanged.
******/
void *
LIST_retrieveList( list_t * this )
{
  dLink * current = this->current;
  void * content = 0;

  if (current)
  {
    if (this->listSize == 1)
    {
      this->top = this->bottom = this->current = 0;
    }
    else if (current == this->top)
    {
      this->top = this->current = current->next;
      this->current->prior = 0;
    }
    else if (current == this->bottom)
    {
      this->bottom = this->current = current->prior;
      this->current->next = 0;
    }
    else
    {
      this->current = current->prior;
      this->current->next = current->next;
      current->next->prior = this->current;
    }
    this->listSize --;
    content = current->content;
    free (current);
  }

  return content;	
}


/* LIST_freeList - Free header, nodes, and content of a list */
void
LIST_freeList( list_t * this )
{
  void * content;

  if (this)
  {
    LIST_topList (this);
    while (this->listSize)
    {
      content=LIST_retrieveList(this);
      this->freeContent (content);
    }
    free (this);
  }
}



/*----------------------------------------------------------------------
    Miscellaneous compare and free functions for doubly linked lists 

These functions are node comparison and freeing functions for doubly
linked lists, for some common types of content.  The compare functions
compare two values via pointers and return -1, 0, +1 to indicate the
relative order.  The free function frees the contents of a node, or
just pretends to if the contents are permanently allocated.
----------------------------------------------------------------------*/


/* floatcmp - compare floats in a doubly linked list */
int
floatcmp( float * f1,
	  float * f2 )
{
  if (*f1 < *f2)
    return -1;
  else if (*f1 > *f2)
    return 1;
  else
    return 0;
}


/* intcmp - compare ints in a doubly linked list */
int
intcmp( int * i1,
	int * i2 )
{
  if (*i1 < *i2)
    return -1;
  else if (*i1 > *i2)
    return 1;
  else
    return 0;
}


/* defaultFreeContent - free contents of a list node */
void
defaultFreeContent( void * content )
{
  free(content);
}


/* nullFreeContent - pretend to free contents of a list node */
void
nullFreeContent( void * content )
{
}
